@@ -1,6 +1,4 @@
 /*
- * $Id: fbpict.c,v 1.4 2005/08/22 03:49:47 vektor Exp $
- *
  * Copyright © 2000 SuSE, Inc.
  *
  * Permission to use, copy, modify, distribute, and sell this software and its
@@ -74,20 +72,6 @@ fbIn (CARD32 x, CARD8 y)
     return m|n|o|p;
 }
 
-static CARD32
-fbIn24 (CARD32 x, CARD8 y)
-{
-    CARD16  a = y;
-    CARD16  t;
-    CARD32  m,n,o,p;
-
-    m = FbInU(x,0,a,t);
-    n = FbInU(x,8,a,t);
-    o = FbInU(x,16,a,t);
-    p = (y << 24);
-    return m|n|o|p;
-}
-
 #define genericCombine24(a,b,c,d) (((a)*(c)+(b)*(d)))
 
 /*
@@ -169,7 +153,7 @@ fbCompositeSolidMask_nx8x8888 (pixman_operator_t   op,
     FbStride	dstStride, maskStride;
     CARD16	w;
 
-    fbComposeGetSolid(pSrc, src);
+    fbComposeGetSolid(pSrc, pDst, src);
 
     dstMask = FbFullMask (pDst->pDrawable->depth);
     srca = src >> 24;
@@ -228,7 +212,7 @@ fbCompositeSolidMask_nx8888x8888C (pixman_operator_t   op,
     CARD16	w;
     CARD32	m, n, o, p;
 
-    fbComposeGetSolid(pSrc, src);
+    fbComposeGetSolid(pSrc, pDst, src);
 
     dstMask = FbFullMask (pDst->pDrawable->depth);
     srca = src >> 24;
@@ -302,7 +286,7 @@ fbCompositeSolidMask_nx8x0888 (pixman_operator_t   op,
     CARD16	w;
 	CARD32 rs,gs,bs,rd,gd,bd;
 
-    fbComposeGetSolid(pSrc, src);
+    fbComposeGetSolid(pSrc, pDst, src);
 
     srca = src >> 24;
     srcia = 255-srca;
@@ -398,7 +382,7 @@ fbCompositeSolidMask_nx8x0565 (pixman_operator_t      op,
     FbStride	dstStride, maskStride;
     CARD16	w,src16;
 
-    fbComposeGetSolid(pSrc, src);
+    fbComposeGetSolid(pSrc, pDst, src);
 
     if (src == 0)
 	return;
@@ -476,7 +460,7 @@ fbCompositeSolidMask_nx8888x0565 (pixman_operator_t      op,
     FbStride	dstStride, maskStride;
     CARD16	w, src16;
 
-    fbComposeGetSolid(pSrc, src);
+    fbComposeGetSolid(pSrc, pDst, src);
 
     if (src == 0)
 	return;
@@ -555,7 +539,7 @@ fbCompositeSolidMask_nx8888x0565C (pixman_operator_t      op,
     CARD16	w;
     CARD32	m, n, o;
 
-    fbComposeGetSolid(pSrc, src);
+    fbComposeGetSolid(pSrc, pDst, src);
 
     srca = src >> 24;
     if (src == 0)
@@ -752,8 +736,6 @@ fbCompositeSrc_8888x0565 (pixman_operator_t  op,
     }
 }
 
-
-
 static void
 fbCompositeSrcAdd_8000x8000 (pixman_operator_t	  op,
 			     PicturePtr pSrc,
@@ -926,7 +908,7 @@ fbCompositeSolidMask_nx1xn (pixman_operator_t   op,
     int		maskXoff, maskYoff;
     FbBits	src;
 
-    fbComposeGetSolid(pSrc, src);
+    fbComposeGetSolid(pSrc, pDst, src);
 
     if ((src & 0xff000000) != 0xff000000)
     {
@@ -1008,7 +990,7 @@ fbCompositeTrans_0565xnx0565(pixman_operator_t      op,
     CARD16	s_16, d_16;
     CARD32	s_32, d_32;
 
-    fbComposeGetSolid (pMask, mask);
+    fbComposeGetSolid (pMask, pDst, mask);
     maskAlpha = mask >> 27;
 
     if (!maskAlpha)
@@ -1085,8 +1067,6 @@ fbCompositeTrans_0565xnx0565(pixman_operator_t      op,
 	}
 }
 
-
-
 /* macros for "i can't believe it's not fast" packed pixel handling */
 #define alphamaskCombine24(a,b) genericCombine24(a,b,maskAlpha,maskiAlpha)
 static void
@@ -1110,7 +1090,7 @@ fbCompositeTrans_0888xnx0888(pixman_operator_t      op,
     FbBits	mask;
     CARD16	maskAlpha,maskiAlpha;
 
-    fbComposeGetSolid (pMask, mask);
+    fbComposeGetSolid (pMask, pDst, mask);
     maskAlpha = mask >> 24;
 	maskiAlpha= 255-maskAlpha;
 
@@ -1151,7 +1131,7 @@ fbCompositeTrans_0888xnx0888(pixman_operator_t      op,
 				setupPackedReader(ws,wt,isrc,wsrc,workingSource);
 
 				/* get to word aligned */
-				switch(!(long)dst&3)
+				switch(~(long)dst&3)
 				{
 					case 1:
 						readPackedSource(rs);
@@ -1227,7 +1207,7 @@ fbCompositeTrans_0888xnx0888(pixman_operator_t      op,
 				srcLine += srcStride;
 				w = width*3;
 				/* get to word aligned */
-				switch(!(long)src&3)
+				switch(~(long)src&3)
 				{
 					case 1:
 						rd=alphamaskCombine24(*src++, *dst)>>8;
@@ -1360,7 +1340,7 @@ pixman_composite (pixman_operator_t	op,
     pixman_region16_t	    *region;
     int		    n;
     pixman_box16_t    *pbox;
-    CompositeFunc   func = 0;
+    CompositeFunc   func = NULL;
     Bool	    srcRepeat = pSrc->pDrawable && pSrc->repeat == RepeatNormal;
     Bool	    maskRepeat = FALSE;
     Bool	    srcTransform = pSrc->transform != 0;
@@ -1369,7 +1349,7 @@ pixman_composite (pixman_operator_t	op,
     Bool	    maskAlphaMap = FALSE;
     Bool	    dstAlphaMap = pDst->alphaMap != 0;
     int		    x_msk, y_msk, x_src, y_src, x_dst, y_dst;
-    int		    w, h, w_this, h_this;
+    unsigned int    w, h, w_this, h_this;
 
 #ifdef USE_MMX
     static Bool mmx_setup = FALSE;
@@ -1378,7 +1358,7 @@ pixman_composite (pixman_operator_t	op,
         mmx_setup = TRUE;
     }
 #endif
-        
+
     xDst += pDst->pDrawable->x;
     yDst += pDst->pDrawable->y;
     if (pSrc->pDrawable) {
@@ -1818,6 +1798,20 @@ pixman_composite (pixman_operator_t	op,
 	    }
 	}
 	break;
+    case PIXMAN_OPERATOR_CLEAR:
+    case PIXMAN_OPERATOR_DST:
+    case PIXMAN_OPERATOR_OVER_REVERSE:
+    case PIXMAN_OPERATOR_IN:
+    case PIXMAN_OPERATOR_IN_REVERSE:
+    case PIXMAN_OPERATOR_OUT:
+    case PIXMAN_OPERATOR_OUT_REVERSE:
+    case PIXMAN_OPERATOR_ATOP:
+    case PIXMAN_OPERATOR_ATOP_REVERSE:
+    case PIXMAN_OPERATOR_XOR:
+    case PIXMAN_OPERATOR_SATURATE:
+    default:
+	/* For any operator not specifically handled above we default out to the general code. */
+	func = NULL;
     }
 
     if (!func) {
@@ -1834,7 +1828,7 @@ pixman_composite (pixman_operator_t	op,
 
     region = pixman_region_create();
     pixman_region_union_rect (region, region, xDst, yDst, width, height);
-    
+
     if (!FbComputeCompositeRegion (region,
 				   pSrc,
 				   pMask,
@@ -1848,7 +1842,7 @@ pixman_composite (pixman_operator_t	op,
 				   width,
 				   height))
 	return;
-    
+
     n = pixman_region_num_rects (region);
     pbox = pixman_region_rects (region);
     while (n--)
@@ -1909,4 +1903,121 @@ pixman_composite (pixman_operator_t	op,
     pixman_region_destroy (region);
 }
 slim_hidden_def(pixman_composite);
+
+/* The CPU detection code needs to be in a file not compiled with
+ * "-mmmx -msse", as gcc would generate CMOV instructions otherwise
+ * that would lead to SIGILL instructions on old CPUs that don't have
+ * it.
+ */
+#if defined(USE_MMX) && !defined(__amd64__) && !defined(__x86_64__)
+
+enum CPUFeatures {
+    NoFeatures = 0,
+    MMX = 0x1,
+    MMX_Extensions = 0x2,
+    SSE = 0x6,
+    SSE2 = 0x8,
+    CMOV = 0x10
+};
+
+static unsigned int detectCPUFeatures(void) {
+    unsigned int result, features;
+    char vendor[13];
+    vendor[0] = 0;
+    vendor[12] = 0;
+    /* see p. 118 of amd64 instruction set manual Vol3 */
+    /* We need to be careful about the handling of %ebx and
+     * %esp here. We can't declare either one as clobbered
+     * since they are special registers (%ebx is the "PIC
+     * register" holding an offset to global data, %esp the
+     * stack pointer), so we need to make sure they have their
+     * original values when we access the output operands.
+     */
+    __asm__ ("pushf\n"
+             "pop %%eax\n"
+             "mov %%eax, %%ecx\n"
+             "xor $0x00200000, %%eax\n"
+             "push %%eax\n"
+             "popf\n"
+             "pushf\n"
+             "pop %%eax\n"
+             "mov $0x0, %%edx\n"
+             "xor %%ecx, %%eax\n"
+             "jz 1f\n"
+
+             "mov $0x00000000, %%eax\n"
+	     "push %%ebx\n"
+             "cpuid\n"
+             "mov %%ebx, %%eax\n"
+	     "pop %%ebx\n"
+             "mov %%eax, %1\n"
+             "mov %%edx, %2\n"
+             "mov %%ecx, %3\n"
+             "mov $0x00000001, %%eax\n"
+	     "push %%ebx\n"
+             "cpuid\n"
+	     "pop %%ebx\n"
+             "1:\n"
+             "mov %%edx, %0\n"
+             : "=r" (result),
+               "=m" (vendor[0]),
+               "=m" (vendor[4]),
+               "=m" (vendor[8])
+             :
+             : "%eax", "%ecx", "%edx"
+        );
+
+    features = 0;
+    if (result) {
+        /* result now contains the standard feature bits */
+        if (result & (1 << 15))
+            features |= CMOV;
+        if (result & (1 << 23))
+            features |= MMX;
+        if (result & (1 << 25))
+            features |= SSE;
+        if (result & (1 << 26))
+            features |= SSE2;
+        if ((result & MMX) && !(result & SSE) && (strcmp(vendor, "AuthenticAMD") == 0)) {
+            /* check for AMD MMX extensions */
+
+            unsigned int result;
+            __asm__("push %%ebx\n"
+                    "mov $0x80000000, %%eax\n"
+                    "cpuid\n"
+                    "xor %%edx, %%edx\n"
+                    "cmp $0x1, %%eax\n"
+                    "jge 1f\n"
+                    "mov $0x80000001, %%eax\n"
+                    "cpuid\n"
+                    "1:\n"
+                    "pop %%ebx\n"
+                    "mov %%edx, %0\n"
+                    : "=r" (result)
+                    :
+                    : "%eax", "%ecx", "%edx"
+                );
+            if (result & (1<<22))
+                features |= MMX_Extensions;
+        }
+    }
+    return features;
+}
+
+Bool
+fbHaveMMX (void)
+{
+    static Bool initialized = FALSE;
+    static Bool mmx_present;
+
+    if (!initialized)
+    {
+        unsigned int features = detectCPUFeatures();
+	mmx_present = (features & (MMX|MMX_Extensions)) == (MMX|MMX_Extensions);
+        initialized = TRUE;
+    }
+
+    return mmx_present;
+}
+#endif /* USE_MMX && !amd64 */
 #endif /* RENDER */
