@@ -869,7 +869,7 @@ gdip_read_bmp_image_from_file_stream (void *pointer, GpImage **image, bool useFi
 		case Format24bppRgb: result->active_bitmap->stride = result->active_bitmap->width * 4;		break;
 		default:
 			/* For other types, we assume 32 bit and translate into 32 bit from source format */
-			result->active_bitmap->pixel_format = Format32bppArgb;
+			result->active_bitmap->pixel_format = Format32bppRgb;
 			result->active_bitmap->stride = result->active_bitmap->width * 4;
 			break;
 	}
@@ -1196,15 +1196,9 @@ gdip_save_bmp_image_to_file_stream (void *pointer, GpImage *image, bool useFile)
 			ptr = current_line;
 			for (k = 0; k < width; k++) {
 				uint32_t color = *iptr++;
-#ifndef WORDS_BIGENDIAN
 				*ptr++ = (color & 0x000000ff);
 				*ptr++ = ((color & 0x0000ff00) >> 8);
 				*ptr++ = ((color & 0x00ff0000) >> 16);
-#else
-				*ptr++ = ((color & 0xff000000) >> 24);
-				*ptr++ = ((color & 0x00ff0000) >> 16);
-				*ptr++ = ((color & 0x0000ff00) >> 8);
-#endif
 			}
 			gdip_write_bmp_data (pointer, current_line, mystride, useFile);
 		}
