@@ -50,12 +50,6 @@
 		((unsigned char *)(pixel))[index+2] = g; \
 		((unsigned char *)(pixel))[index+3] = b; \
 	} while (0)
-#define get_pixel_bgra(color, b, g, r, a) do { \
-		a = (color & 0x000000ff); \
-		r = (color & 0x0000ff00) >> 8; \
-		g = (color & 0x00ff0000) >> 16; \
-		b = (color & 0xff000000) >> 24; \
-	} while (0)
 #else
 #define set_pixel_bgra(pixel,index,b,g,r,a) do { \
 		((unsigned char *)(pixel))[index+0] = b; \
@@ -63,13 +57,13 @@
 		((unsigned char *)(pixel))[index+2] = r; \
 		((unsigned char *)(pixel))[index+3] = a; \
 	} while (0)
+#endif
 #define get_pixel_bgra(color, b, g, r, a) do { \
 		a = ((color & 0xff000000) >> 24); \
 		r = ((color & 0x00ff0000) >> 16); \
 		g = ((color & 0x0000ff00) >> 8); \
 		b = (color & 0x000000ff); \
 	} while(0)
-#endif
 
 #ifdef CAIRO_HAS_XLIB_SURFACE
 #ifdef USE_INCLUDED_CAIRO
@@ -107,6 +101,7 @@ typedef gpointer HANDLE;
 typedef gpointer *LPHANDLE;
 typedef guint32 SOCKET;
 typedef gpointer HMODULE;
+typedef gpointer HRGN;
 #define CONST const
 #define VOID void
 
@@ -142,6 +137,7 @@ typedef ImageAbort GetThumbnailImageAbort;
 #define GDIPCONST const
 #define CURVE_MIN_TERMS 1
 #define CURVE_MAX_TERMS 7
+#define DEFAULT_TEXT_CONTRAST	4
 
 #define CAIRO_AA_OFFSET_X	1
 #define CAIRO_AA_OFFSET_Y	0.5
@@ -341,6 +337,7 @@ typedef enum {
 	WMF,
 	EMF,
 	ICON,
+	MEMBMP,
 	INVALID
 } ImageFormat;
 
@@ -603,9 +600,7 @@ typedef enum {
 	ColorChannelFlagsLast
 } GpColorChannelFlags;
 
-typedef struct {
-	int	unused;
-} GpGraphicsContainer;
+typedef unsigned int GpGraphicsContainer;
 
 /* private enum */
 
@@ -634,24 +629,9 @@ typedef enum {
  */
 
 #define GBD_OWN_SCAN0	(1<<8)
-#define GBD_READ_ONLY	(1<<9)
+#define GBD_WRITE_OK	(1<<9)
 #define GBD_LOCKED	(1<<10)
 
-typedef struct {    /* Keep in sync with BitmapData.cs */
-	unsigned int Width;
-	unsigned int Height;
-	int          Stride;
-	int          PixelFormat;
-	byte* 	     Scan0;
-	unsigned int Reserved;
-	/*Added to keep track of position for displaying next frame
-	in a gif*/
-	unsigned int Top; 
-	unsigned int Left;
-	int	     ByteCount;
-	char*	     Bytes;
-} GdipBitmapData, BitmapData;
-
 typedef struct {
 	int X, Y, Width, Height;
 } GpRect, Rect;
@@ -687,6 +667,7 @@ typedef struct {
         GpDashStyle dash_style;
 	GpLineCap line_cap;  /* Cairo supports only same kind of end caps for both the ends. */
 	GpLineCap end_cap;
+	GpDashCap dash_cap;  /* not supported */
 	int compound_count;
 	float *compound_array;
         GpPenAlignment mode;
@@ -695,7 +676,7 @@ typedef struct {
 	BOOL own_dash_array; /* flag to mark if pen maintains its own array or global array */
         float *dash_array;
         GpUnit unit; /* Always set to UnitWorld. */
-        GpMatrix *matrix;
+	cairo_matrix_t matrix;
         BOOL changed; /* flag to mark if pen is changed and needs setup */
 } GpPen;
 
@@ -748,6 +729,7 @@ typedef struct {
  
 typedef struct {
 	cairo_matrix_t		matrix;
+	cairo_matrix_t		previous_matrix;
 	GpRegion*		clip;
 	cairo_matrix_t		clip_matrix;
 	GpCompositingMode    	composite_mode;
@@ -760,11 +742,13 @@ typedef struct {
 	PixelOffsetMode 	pixel_mode;
 	int			org_x;
 	int			org_y;
+	int			text_contrast;
 } GpState;
 
 typedef struct {
 	cairo_t         *ct;
 	GpMatrix	*copy_of_ctm;
+	cairo_matrix_t	previous_matrix;
 	Display		*display;
 	Drawable	drawable;
 	void            *image;
@@ -790,6 +774,7 @@ typedef struct {
 	int		render_origin_y;
 	float		dpi_x;
 	float		dpi_y;
+	int		text_contrast;
 } GpGraphics;
 
 
@@ -826,33 +811,56 @@ typedef struct {
 	VOID*	value;
 } PropertyItem;
 
+/* This structure is mirrored in System.Drawing.Imaging.BitmapData.
+   Any changes here must also be made to BitmapData.cs */
 typedef struct {
-	GUID frameDimension;
-	int count;
-	BitmapData *frames;
-} FrameInfo;
+	unsigned int	width;
+	unsigned int	height;
+	int		stride;
+	int		pixel_format;
+	byte 		*scan0;
+	unsigned int	reserved;
+	ColorPalette	*palette;
+
+	int		property_count;		/* Number of properties */
+	PropertyItem 	*property;		/* Properties associated with image */
+
+	float 		dpi_horz;		/* */
+	float 		dpi_vert;		/* */
+	ImageFlags	image_flags;		/* Alpha, ColorSpace, etc. */
+
+	unsigned int	left;			/* left display coordinate of frame */
+	unsigned int	top;			/* top display coordinate of frame */
+	unsigned int	x;			/* LockBits: left coordinate of locked rectangle */
+	unsigned int	y;			/* LockBits: top coordinate of locked rectangle */
+
+	int		transparent;		/* Index of transparent color (<24bit only) */
+} GdipBitmapData, BitmapData;
 
 typedef struct {
-	ImageType     	type;
-	cairo_surface_t *surface;
-	int 		imageFlags;
-	int 		height;
-	int 		width;
-	float 		horizontalResolution;
-	ColorPalette 	*palette;
-	int	 	pixFormat;
-	PropertyItem 	*propItems;
-	float 		verticalResolution;
-	ImageFormat     format;
-	int		frameDimensionCount;
-	FrameInfo	*frameDimensionList;
-} GpImage;
+	int		count;			/* Number of bitmaps contained in this frame */
+	BitmapData	*bitmap;		/* Bitmaps for this frame */
+	GUID		frame_dimension;	/* GUID describing the frame type */
+} FrameData;
 
 typedef struct {
-	GpImage		image;
+	/* Image Description */
+	ImageType     	type;			/* Undefined, Bitmap, MetaFile */
+	ImageFormat     image_format;		/* BMP, TIF, GIF, PNG, etc.	*/
+
+	/* Image Data */
+	int		num_of_frames;		/* Number of frames */
+	FrameData	*frames;		/* Array of frames (Page, Time, Resolution) for the image */
+
+	/* Tracking of active image */
+	int		active_frame;		/* Index of frame currently used */
+	int		active_bitmap_no;	/* Index of active bitmap in current frame */
+	BitmapData	*active_bitmap;		/* Pointer to active frame/bitmap; DO NOT free() */
+
+	/* Internal fields */
         int             cairo_format;
-	BitmapData	data;
-} GpBitmap;
+	cairo_surface_t *surface;
+} GpImage, GpBitmap;
 
 typedef struct {
         FcFontSet*  fontset;
@@ -860,7 +868,6 @@ typedef struct {
         FcConfig*   config;
 } GpFontCollection;
 
-
 typedef struct {
         FcPattern*	pattern;
 	bool		allocated;
@@ -879,6 +886,7 @@ typedef struct {
 	GpFontFamily	    *family;
 	float               emSize;
 	GpUnit              unit;
+	cairo_t*            ct;
 } GpFont;
 
 
@@ -991,15 +999,16 @@ typedef struct {
 
 typedef struct
 {
-	Rect region;
-	int x, y;               /* the offset of the next byte that will be loaded, once the buffer is depleted */
-	unsigned short buffer;
-	int p;                  /* index of pixel within 'buffer' that was returned by the last call to gdip_pixel_stream_get_next () */
-	int one_pixel_mask, one_pixel_shift;
-	int pixels_per_byte;	/* a negative value is used to indicate a count of bytes per pixel for depths of more than 8 bits */
+	Rect		region;
+	int		x, y;			/* the offset of the next byte that will be loaded, once the buffer is depleted */
+	unsigned short	buffer;
+	int		p;			/* index of pixel within 'buffer' that was returned by the last call to gdip_pixel_stream_get_next () */
+	int		one_pixel_mask;
+	int		one_pixel_shift;
+	int		pixels_per_byte;	/* a negative value is used to indicate a count of bytes per pixel for depths of more than 8 bits */
 
-	BitmapData *data;
-	unsigned char *scan;
+	BitmapData	*data;
+	unsigned char	*scan;
 } StreamingState;
 
 GpStatus gdip_init_pixel_stream (StreamingState *state, BitmapData *data, int x, int y, int w, int h);
@@ -1013,10 +1022,31 @@ void gdip_pixel_stream_set_next (StreamingState *state, unsigned int pixel_value
  */
 void gdip_image_init              (GpImage *image);
 
-void gdip_bitmap_init  (GpBitmap *bitmap);
-GpBitmap *gdip_bitmap_new   (void);
-void gdip_bitmap_dispose (GpBitmap *bitmap);
-GpStatus gdip_bitmap_clone (GpBitmap *bitmap, GpBitmap **clonedbitmap);
+void		gdip_bitmap_init(GpBitmap *bitmap);
+GpBitmap	*gdip_bitmap_new(void);
+GpBitmap	*gdip_bitmap_new_with_frame(const GUID *dimension, bool add_bitmapdata);
+FrameData	*gdip_frame_add(GpBitmap *bitmap, const GUID *dimension);
+BitmapData	*gdip_frame_add_bitmapdata(FrameData *frame);
+void		gdip_bitmap_dispose (GpBitmap *bitmap);
+GpStatus	gdip_bitmap_clone (GpBitmap *bitmap, GpBitmap **clonedbitmap);
+GpStatus	gdip_bitmap_setactive(GpBitmap *bitmap, const GUID *dimension, int index);
+GpStatus	gdip_bitmapdata_clone(BitmapData *src, BitmapData **dest, int count);
+ColorPalette	*gdip_palette_clone(ColorPalette *original);
+GpStatus	gdip_property_get_short(int offset, void *value, unsigned short *result);
+GpStatus	gdip_property_get_long(int offset, void *value, guint32 *result);
+GpStatus	gdip_property_get_srational(int offset, void *value, unsigned short *numerator, unsigned short *denominator);
+GpStatus	gdip_property_get_rational(int offset, void *value, guint32 *numerator, guint32 *denominator);
+GpStatus	gdip_bitmapdata_property_add(BitmapData *bitmap_data, PROPID id, ULONG length, WORD type, VOID *value);
+GpStatus	gdip_bitmapdata_property_add_byte(BitmapData *bitmap_data, PROPID id, byte value);
+GpStatus	gdip_bitmapdata_property_add_short(BitmapData *bitmap_data, PROPID id, unsigned short value);
+GpStatus	gdip_bitmapdata_property_add_long(BitmapData *bitmap_data, PROPID id, guint32 value);
+GpStatus	gdip_bitmapdata_property_add_ASCII(BitmapData *bitmap_data, PROPID id, unsigned char *value);
+GpStatus	gdip_bitmapdata_property_add_rational(BitmapData *bitmap_data, PROPID id, guint32 numerator, guint32 denominator);
+GpStatus	gdip_bitmapdata_property_add_srational(BitmapData *bitmap_data, PROPID id, unsigned short numerator, unsigned short denominator);
+GpStatus	gdip_bitmapdata_property_remove_id(BitmapData *bitmap_data, PROPID id);
+GpStatus	gdip_bitmapdata_property_remove_index(BitmapData *bitmap_data, int index);
+GpStatus	gdip_bitmapdata_property_find_id(BitmapData *bitmap_data, PROPID id, int *index);
+
 
 void gdip_graphics_init (GpGraphics *graphics, cairo_surface_t *surface);
 GpGraphics *gdip_graphics_new (cairo_surface_t *surface);
@@ -1049,6 +1079,9 @@ GpStatus GdipBitmapGetPixel ( GpBitmap *bmp, int x, int y, ARGB *color);
 
 
 /* Graphics */
+GpStatus GdipBeginContainer (GpGraphics *graphics, GDIPCONST GpRectF* dstrect, GDIPCONST GpRectF *srcrect, GpUnit unit, GpGraphicsContainer *state);
+GpStatus GdipBeginContainer2 (GpGraphics *graphics, GpGraphicsContainer* state);
+GpStatus GdipBeginContainerI (GpGraphics *graphics, GDIPCONST GpRect* dstrect, GDIPCONST GpRect *srcrect, GpUnit unit, GpGraphicsContainer *state);
 GpStatus GdipCreateFromHDC (void *hDC, GpGraphics **graphics);
 GpStatus GdipDeleteGraphics (GpGraphics *graphics);
 GpStatus GdipGetDC (GpGraphics *graphics, void **hDC);
@@ -1065,8 +1098,8 @@ GpStatus GdipDrawCurve (GpGraphics *graphics, GpPen *pen, GpPointF *points, int
 GpStatus GdipDrawCurveI (GpGraphics *graphics, GpPen *pen, GpPoint *points, int count);
 GpStatus GdipDrawCurve2 (GpGraphics *graphics, GpPen *pen, GpPointF *points, int count, float tension);
 GpStatus GdipDrawCurve2I (GpGraphics *graphics, GpPen *pen, GpPoint *points, int count, float tension);
-GpStatus GdipDrawCurve3 (GpGraphics *graphics, GpPen *pen, GpPointF *points, int count, int offset, float numOfSegments, float tension);
-GpStatus GdipDrawCurve3I (GpGraphics *graphics, GpPen *pen, GpPoint *points, int count, int offset, float numOfSegments, float tension);
+GpStatus GdipDrawCurve3 (GpGraphics *graphics, GpPen *pen, GpPointF *points, int count, int offset, int numOfSegments, float tension);
+GpStatus GdipDrawCurve3I (GpGraphics *graphics, GpPen *pen, GpPoint *points, int count, int offset, int numOfSegments, float tension);
 GpStatus GdipDrawClosedCurve (GpGraphics *graphics, GpPen *pen, GpPointF *points, int count);
 GpStatus GdipDrawClosedCurveI (GpGraphics *graphics, GpPen *pen, GpPoint *points, int count);
 GpStatus GdipDrawClosedCurve2 (GpGraphics *graphics, GpPen *pen, GpPointF *points, int count, float tension);
@@ -1129,6 +1162,13 @@ GpStatus GdipSetTextContrast(GpGraphics *graphics, UINT contrast);
 GpStatus GdipGetTextContrast(GpGraphics *graphics, UINT *contrast);
 GpStatus GdipSetSmoothingMode(GpGraphics *graphics, SmoothingMode mode);
 GpStatus GdipGetSmoothingMode(GpGraphics *graphics, SmoothingMode *mode);
+GpStatus GdipResetClip (GpGraphics *graphics);
+GpStatus GdipSetClipGraphics (GpGraphics *graphics, GpGraphics *srcgraphics, CombineMode combineMode);
+GpStatus GdipSetClipHrgn (GpGraphics *graphics, void *hRgn, CombineMode combineMode);
+GpStatus GdipSetClipPath (GpGraphics *graphics, GpPath *path, CombineMode combineMode);
+GpStatus GdipSetClipRect (GpGraphics *graphics, float x, float y, float width, float height, CombineMode combineMode);
+GpStatus GdipSetClipRectI (GpGraphics *graphics, int x, int y, int width, int height, CombineMode combineMode);
+GpStatus GdipSetClipRegion (GpGraphics *graphics, GpRegion *region, CombineMode combineMode);
 GpStatus GdipSetVisibleClip_linux (GpGraphics *graphics, GpRect *rect);
 GpStatus GdipTranslateClip (GpGraphics *graphics, float dx, float dy);
 GpStatus GdipTranslateClipI (GpGraphics *graphics, int dx, int dy);
@@ -1303,8 +1343,8 @@ GpStatus GdipSetImageAttributesOutputChannelColorProfile(GpImageAttributes *imag
 GpStatus GdipSetImageAttributesRemapTable(GpImageAttributes *imageattr, ColorAdjustType type, BOOL enableFlag, UINT mapSize, GDIPCONST GpColorMap *map);
 GpStatus GdipSetImageAttributesWrapMode(GpImageAttributes *imageattr, WrapMode wrap, ARGB argb, BOOL clamp); 
 GpStatus GdipGetImageAttributesAdjustedPalette(GpImageAttributes *imageattr, ColorPalette *colorPalette, ColorAdjustType type); 
-GpStatus GdipSetImageAttributesColorMatrix(GpImageAttributes *imageattr, ColorAdjustType type, BOOL enableFlag,  GpColorMatrix* colorMatrix,
-	GpColorMatrix* grayMatrix, GpColorMatrixFlags flags);
+GpStatus GdipSetImageAttributesColorMatrix(GpImageAttributes *imageattr, ColorAdjustType type, BOOL enableFlag, 
+	GDIPCONST GpColorMatrix* colorMatrix, GDIPCONST GpColorMatrix* grayMatrix, GpColorMatrixFlags flags);
 GpStatus GdipSetImageAttributesOutputChannel(GpImageAttributes *imageattr, ColorAdjustType type, BOOL enableFlag, GpColorChannelFlags channelFlags);
 
 /* Region */
@@ -1340,9 +1380,9 @@ GpStatus GdipTransformRegion(GpRegion *region, GpMatrix *matrix);
 
 
 /* Encoders / Decoders */
-GpStatus GdipGetImageDecodersSize (int *numDecoders, int *size);
-GpStatus GdipGetImageDecoders (int numDecoders, int size, ImageCodecInfo *decoders);
-GpStatus GdipGetImageEncodersSize (int *numEncoders, int *size);
+GpStatus GdipGetImageDecodersSize (UINT *numDecoders, UINT *size);
+GpStatus GdipGetImageDecoders (UINT numDecoders, UINT size, ImageCodecInfo *decoders);
+GpStatus GdipGetImageEncodersSize (UINT *numEncoders, UINT *size);
 GpStatus GdipGetImageEncoders (UINT numEncoders, UINT size, ImageCodecInfo *encoders);
 
 
@@ -1357,6 +1397,7 @@ void gdip_calculate_coefficients (int count, int terms, float **coefficients, in
 
 /* Memory */
 void *GdipAlloc (int size);
+void *GdipRealloc (void *org, int size);
 void *GdipCalloc (size_t nelem, size_t elsize); 
 void GdipFree (void *ptr);
 
@@ -1365,7 +1406,7 @@ int fcmp (double x1, double x2, double epsilon);
 float gdip_get_display_dpi();
 float gdip_unitx_convgr (GpGraphics *graphics, float nSrc);
 float gdip_unity_convgr (GpGraphics *graphics, float nSrc);
-void gdip_unit_conversion (Unit from, Unit to, float dpi, GraphicsType type, float nSrc, float* nTrg);
+float gdip_unit_conversion (Unit from, Unit to, float dpi, GraphicsType type, float nSrc);
 
 cairo_content_t from_cairoformat_to_content (cairo_format_t format);
 
@@ -1382,6 +1423,10 @@ bool utf8_to_ucs2(const gchar *utf8, gunichar2 *ucs2, int ucs2_len);
 int utf8_encode_ucs2char(gunichar2 unichar, unsigned char *dest);
 int utf8_decode_ucs2char(const unsigned char *src, gunichar2 *uchar);
 
+GpStatus gdip_get_pattern_status (cairo_pattern_t *pat);
+
+void gdip_font_clear_pattern_cache (void);
+
 
 /* Stream handling bits */
 typedef int (*GetHeaderDelegate) (unsigned char *, int);
