@@ -21,7 +21,7 @@ gdip_load_gif_image_from_stream_delegate (GetBytesDelegate getBytesFunc,
                                            GpImage **image);
 					   
 GpStatus 
-gdip_save_gif_image_to_file (char *filename, GpImage *image);
+gdip_save_gif_image_to_file (unsigned char *filename, GpImage *image);
 
 
 GpStatus
