@@ -77,13 +77,13 @@ gdip_process_bitmap_attributes (GpBitmap *bitmap, void **dest, GpImageAttributes
 
 	GpImageAttribute *imgattr, *def;
 	GpImageAttribute *colormap, *gamma, *trans, *cmatrix;
-	void *scan0;
-	GpBitmap bmpdest;
+	GpBitmap *bmpdest;
 	int x,y, cnt;
 	ARGB color;	
 	byte *color_p = (byte *)&color;
 	
 	*allocated = FALSE;
+	bmpdest = NULL;
 	
 	if (!bitmap || !dest || !attr) 
 		return;
@@ -120,12 +120,12 @@ gdip_process_bitmap_attributes (GpBitmap *bitmap, void **dest, GpImageAttributes
 
 	if (colormap->colormap_elem || gamma->gamma_correction || trans->key_enabled || 
 	    (cmatrix->colormatrix_enabled && cmatrix->colormatrix != NULL)) {
-		scan0 = GdipAlloc (bitmap->data.Stride * bitmap->data.Height);
-		memcpy (scan0, bitmap->data.Scan0, bitmap->data.Stride * bitmap->data.Height);
-		*dest = scan0;
-		memcpy (&bmpdest, bitmap, sizeof (GpBitmap));
-		bmpdest.data.Bytes = NULL;
-		bmpdest.data.Scan0 = scan0;
+		bitmap->active_bitmap->pixel_format = Format32bppArgb;
+		bmpdest = gdip_bitmap_new_with_frame(NULL, FALSE);
+		gdip_bitmapdata_clone(bitmap->active_bitmap, &bmpdest->frames[0].bitmap, 1);
+		bmpdest->frames[0].count = 1;
+		gdip_bitmap_setactive(bmpdest, NULL, 0);
+		*dest = bmpdest->active_bitmap->scan0;
 		*allocated = TRUE;
 	}
 	
@@ -136,18 +136,18 @@ gdip_process_bitmap_attributes (GpBitmap *bitmap, void **dest, GpImageAttributes
 	
 	/* Color mapping */
 	if (colormap->colormap_elem) {
-		for (y = 0; y <bitmap->data.Height; y++) {	
-			for (x = 0; x <bitmap->data.Width; x++) {
+		for (y = 0; y <bitmap->active_bitmap->height; y++) {	
+			for (x = 0; x <bitmap->active_bitmap->width; x++) {
 				GpColorMap* clrmap = colormap->colormap;
 				int found;
 				
-				GdipBitmapGetPixel (&bmpdest, x, y, &color);
+				GdipBitmapGetPixel (bmpdest, x, y, &color);
 				
 				for (cnt = 0; cnt < colormap->colormap_elem; cnt++, clrmap++) {
 				  
 					if (color == clrmap->oldColor.Color) {						
 						color = clrmap->newColor.Color;						
-						GdipBitmapSetPixel (&bmpdest, x, y, color);
+						GdipBitmapSetPixel (bmpdest, x, y, color);
 						break;
 					}
 				}
@@ -157,12 +157,12 @@ gdip_process_bitmap_attributes (GpBitmap *bitmap, void **dest, GpImageAttributes
 	
 	/* Gamma correction */
 	if (gamma->gamma_correction) {
-		for (y = 0; y <bitmap->data.Height; y++) {	
-			for (x = 0; x <bitmap->data.Width; x++) {
+		for (y = 0; y < bitmap->active_bitmap->height; y++) {	
+			for (x = 0; x < bitmap->active_bitmap->width; x++) {
 			
 				BYTE r,g,b,a;					
 				
-				GdipBitmapGetPixel (&bmpdest, x, y, &color);		
+				GdipBitmapGetPixel (bmpdest, x, y, &color);		
 			
 				a = (color & 0xff000000) >> 24;
 				r = (color & 0x00ff0000) >> 16;
@@ -179,7 +179,7 @@ gdip_process_bitmap_attributes (GpBitmap *bitmap, void **dest, GpImageAttributes
 				
 				color = b | (g  << 8) | (r << 16) | (a << 24);
 					
-				GdipBitmapSetPixel (&bmpdest, x, y, color);
+				GdipBitmapSetPixel (bmpdest, x, y, color);
 			}	
 		}
 		
@@ -187,14 +187,14 @@ gdip_process_bitmap_attributes (GpBitmap *bitmap, void **dest, GpImageAttributes
 	
 	/* Apply transparency range */
 	if (trans->key_enabled) {
-		for (y = 0; y <bitmap->data.Height; y++) {	
-			for (x = 0; x <bitmap->data.Width; x++) {
+		for (y = 0; y < bitmap->active_bitmap->height; y++) {	
+			for (x = 0; x < bitmap->active_bitmap->width; x++) {
 				
-				GdipBitmapGetPixel (&bmpdest, x, y, &color);					
+				GdipBitmapGetPixel (bmpdest, x, y, &color);					
 				
 				if (color >= trans->key_colorlow && color <= trans->key_colorhigh) {
 					color = color & 0x00ffffff; /* Alpha = 0 */
-					GdipBitmapSetPixel (&bmpdest, x, y, color);
+					GdipBitmapSetPixel (bmpdest, x, y, color);
 				}
 			}	
 		}
@@ -203,14 +203,17 @@ gdip_process_bitmap_attributes (GpBitmap *bitmap, void **dest, GpImageAttributes
 
 	/* Apply Color Matrix */
 	if (cmatrix->colormatrix_enabled && cmatrix->colormatrix) {
-		for (y = 0; y <bitmap->data.Height; y++) {	
-			for (x = 0; x <bitmap->data.Width; x++) {
+		for (y = 0; y < bitmap->active_bitmap->height; y++) {	
+			for (x = 0; x < bitmap->active_bitmap->width; x++) {
 
 				byte r,g,b,a;
 				int r_new,g_new,b_new,a_new;
 				
-				GdipBitmapGetPixel (&bmpdest, x, y, &color);
+				GdipBitmapGetPixel (bmpdest, x, y, &color);
 				get_pixel_bgra (color, b, g, r, a);
+				/* Don't transform transparent pixels */
+				if (a == 0)
+					continue;
 
 				r_new = (r * cmatrix->colormatrix->m[0][0] + g * cmatrix->colormatrix->m[1][0] + b * cmatrix->colormatrix->m[2][0] +
 					a * cmatrix->colormatrix->m[3][0] + (255 * cmatrix->colormatrix->m[4][0]));
@@ -230,14 +233,19 @@ gdip_process_bitmap_attributes (GpBitmap *bitmap, void **dest, GpImageAttributes
 				if (a_new > 0xff) a_new = 0xff;
 
 				set_pixel_bgra (color_p, 0, (byte) b_new, (byte) g_new, (byte) r_new, (byte) a_new);
-				GdipBitmapSetPixel (&bmpdest, x, y, color);
+				GdipBitmapSetPixel (bmpdest, x, y, color);
 			}	
 		}
 	
 	}
+
+	if (bmpdest != NULL) {
+		bmpdest->active_bitmap->scan0 = NULL;
+		gdip_bitmap_dispose(bmpdest);
+	}
 }
 
-
+/* coverity[+alloc : arg-*0] */
 GpStatus
 GdipCreateImageAttributes (GpImageAttributes **imageattr)
 {
@@ -247,6 +255,11 @@ GdipCreateImageAttributes (GpImageAttributes **imageattr)
 		return InvalidParameter;
 
 	result = (GpImageAttributes *) GdipAlloc (sizeof (GpImageAttributes));
+	if (!result) {
+		*imageattr = NULL;
+		return OutOfMemory;
+	}
+
 	gdip_init_image_attribute (&result->def);
 	gdip_init_image_attribute (&result->bitmap);
 	gdip_init_image_attribute (&result->brush);
@@ -259,7 +272,7 @@ GdipCreateImageAttributes (GpImageAttributes **imageattr)
 	return Ok;        
 }
 
-
+/* coverity[+alloc : arg-*1] */
 GpStatus
 GdipCloneImageAttributes (GDIPCONST GpImageAttributes *imageattr, GpImageAttributes **cloneImageattr)
 {
@@ -269,6 +282,10 @@ GdipCloneImageAttributes (GDIPCONST GpImageAttributes *imageattr, GpImageAttribu
 		return InvalidParameter;
 
 	result = (GpImageAttributes *) GdipAlloc (sizeof (GpImageAttributes));
+	if (!result) {
+		*cloneImageattr = NULL;
+		return OutOfMemory;
+	}
 
 	memcpy (result, imageattr, sizeof (GpImageAttributes));
 
@@ -344,10 +361,10 @@ GdipSetImageAttributesColorKeys (GpImageAttributes *imageattr, ColorAdjustType t
         ARGB colorLow, ARGB colorHigh)
 {
 	GpImageAttribute *imgattr;
-		
+
 	if (!imageattr)
 		return InvalidParameter;
-		
+
 	imgattr = gdip_get_image_attribute (imageattr, type);
 	
 	if (!imgattr)
@@ -430,14 +447,14 @@ GdipGetImageAttributesAdjustedPalette (GpImageAttributes *imageattr, ColorPalett
 	return NotImplemented;
 }
 
-
+/* MonoTODO - grayMatrix and flags parameters are ignored */
 GpStatus 
-GdipSetImageAttributesColorMatrix (GpImageAttributes *imageattr, ColorAdjustType type, BOOL enableFlag,  GpColorMatrix* colorMatrix,
-	GpColorMatrix* grayMatrix, GpColorMatrixFlags flags)
+GdipSetImageAttributesColorMatrix (GpImageAttributes *imageattr, ColorAdjustType type, BOOL enableFlag, 
+	GDIPCONST GpColorMatrix* colorMatrix, GDIPCONST GpColorMatrix* grayMatrix, GpColorMatrixFlags flags)
 {
 	GpImageAttribute *imgattr;
 	
-	if (!imageattr)
+	if (!imageattr || (!colorMatrix && enableFlag))
 		return InvalidParameter;
 		
 	imgattr = gdip_get_image_attribute (imageattr, type);
@@ -446,8 +463,11 @@ GdipSetImageAttributesColorMatrix (GpImageAttributes *imageattr, ColorAdjustType
 		return InvalidParameter;
 
 	if (colorMatrix) {
-		if (imgattr->colormatrix == NULL)
+		if (!imgattr->colormatrix) {
 			imgattr->colormatrix =  GdipAlloc (sizeof (GpColorMatrix));
+			if (!imgattr->colormatrix)
+				return OutOfMemory;
+		}
 
 		memcpy (imgattr->colormatrix, colorMatrix, sizeof (GpColorMatrix));
 	}
