@@ -47,8 +47,7 @@ gdip_pathgradient_init (GpPathGradient *pg)
 	pg->focusScales.X = 0.0f;
 	pg->focusScales.Y = 0.0f;
 	pg->wrapMode = WrapModeClamp;
-	pg->transform = NULL;
-	GdipCreateMatrix (&pg->transform);
+	cairo_matrix_init_identity (&pg->transform);
 	pg->presetColors = (InterpolationColors *) GdipAlloc (sizeof (InterpolationColors));
 	pg->presetColors->count = 1;
 	pg->presetColors->colors = (ARGB *) GdipAlloc (sizeof (ARGB));
@@ -61,11 +60,10 @@ gdip_pathgradient_init (GpPathGradient *pg)
 	pg->blend->positions = (float *) GdipAlloc (sizeof (float));
 	pg->blend->factors [0] = 1.0;
 	pg->blend->positions[0] = 0.0;
-	pg->rectangle = (GpRectF *) GdipAlloc (sizeof (GpRectF));
-	pg->rectangle->X = 0.0;
-	pg->rectangle->Y = 0.0;
-	pg->rectangle->Width = 0.0;
-	pg->rectangle->Height = 0.0;
+	pg->rectangle.X = 0.0;
+	pg->rectangle.Y = 0.0;
+	pg->rectangle.Width = 0.0;
+	pg->rectangle.Height = 0.0;
 	pg->pattern = NULL;
 }
 
@@ -86,11 +84,12 @@ gdip_pgrad_clone_brush (GpBrush *brush, GpBrush **clonedBrush)
 	GpPathGradient *pgbrush;
 	GpPathGradient *newbrush;
 
-	g_return_val_if_fail (brush != NULL, InvalidParameter);
+	if (!brush || !clonedBrush)
+		return InvalidParameter;
 
 	newbrush = (GpPathGradient *) GdipAlloc (sizeof (GpPathGradient));
-
-	g_return_val_if_fail (newbrush != NULL, OutOfMemory);
+	if (!newbrush)
+		return OutOfMemory;
 
 	pgbrush = (GpPathGradient *) brush;
 
@@ -102,26 +101,28 @@ gdip_pgrad_clone_brush (GpBrush *brush, GpBrush **clonedBrush)
 	}
 
 	newbrush->boundaryColors = GdipAlloc (sizeof(ARGB) * pgbrush->boundaryColorsCount);
+	if (!newbrush->boundaryColors)
+		goto NO_BOUNDARY_COLORS;
+
 	memcpy (newbrush->boundaryColors, pgbrush->boundaryColors, sizeof(ARGB) * pgbrush->boundaryColorsCount);
 	newbrush->boundaryColorsCount = pgbrush->boundaryColorsCount;
 	newbrush->focusScales = pgbrush->focusScales;
 	newbrush->center = pgbrush->center;
 	newbrush->centerColor = pgbrush->centerColor;
 	newbrush->wrapMode = pgbrush->wrapMode;
-	GdipCloneMatrix (pgbrush->transform, &newbrush->transform);
+	gdip_cairo_matrix_copy (&newbrush->transform, &pgbrush->transform);
 
-	if (pgbrush->rectangle) {
-		newbrush->rectangle = (GpRectF *) GdipAlloc (sizeof (GpRectF));
-		memcpy (newbrush->rectangle, pgbrush->rectangle, sizeof (GpRectF));
-	} else {
-		newbrush->rectangle= NULL;
-	}
+	newbrush->rectangle.X = pgbrush->rectangle.X;
+	newbrush->rectangle.Y = pgbrush->rectangle.Y;
+	newbrush->rectangle.Width = pgbrush->rectangle.Width;
+	newbrush->rectangle.Height = pgbrush->rectangle.Height;
 
 	newbrush->presetColors = (InterpolationColors *) GdipAlloc (sizeof (InterpolationColors));
 
 	if (newbrush->presetColors == NULL) 
 		goto NO_PRESET;
 
+	newbrush->presetColors->count = pgbrush->presetColors->count;
 	if (pgbrush->presetColors->count > 0) {
 		newbrush->presetColors->colors = (ARGB *) GdipAlloc (pgbrush->presetColors->count * sizeof (ARGB));
 		if (newbrush->presetColors->colors == NULL) 
@@ -136,9 +137,12 @@ gdip_pgrad_clone_brush (GpBrush *brush, GpBrush **clonedBrush)
 	} else {
 		memcpy (newbrush->presetColors, pgbrush->presetColors, sizeof (InterpolationColors));
 	}
+
 	newbrush->blend = (Blend *) GdipAlloc (sizeof (Blend));
 	if (newbrush->blend == NULL)
 		goto NO_BLEND;
+
+	newbrush->blend->count = pgbrush->blend->count;
 	if (pgbrush->blend->count > 0) {
 		newbrush->blend->factors = (float *) GdipAlloc (pgbrush->blend->count * sizeof (float));
 		if (newbrush->blend->factors == NULL) 
@@ -163,7 +167,8 @@ NO_PRESET_POSITIONS:
 NO_PRESET_COLORS:
 	GdipFree (newbrush->presetColors);
 NO_PRESET:
-	GdipDeleteMatrix (newbrush->transform);
+	GdipFree (newbrush->boundaryColors);
+NO_BOUNDARY_COLORS:
 	GdipFree (newbrush);
 	return OutOfMemory;
 
@@ -182,53 +187,110 @@ gdip_pgrad_destroy (GpBrush *brush)
 {
 	GpPathGradient *pgbrush;
 
-	g_return_val_if_fail (brush != NULL, InvalidParameter);
+	if (!brush)
+		return InvalidParameter;
 
 	pgbrush = (GpPathGradient *) brush;
 
-	if (pgbrush->boundary)
+	if (pgbrush->boundary) {
 		GdipDeletePath (pgbrush->boundary);
-	pgbrush->boundary = NULL;
+		pgbrush->boundary = NULL;
+	}
 
-	if (pgbrush->boundaryColors)
+	if (pgbrush->boundaryColors) {
 		GdipFree (pgbrush->boundaryColors);
-	pgbrush->boundaryColors = NULL;
+		pgbrush->boundaryColors = NULL;
+	}
 
-	if (pgbrush->rectangle)
-		GdipFree (pgbrush->rectangle);
-	pgbrush->rectangle = NULL;
-
-	if (pgbrush->pattern)
+	if (pgbrush->pattern) {
 		cairo_pattern_destroy (pgbrush->pattern);
-	pgbrush->pattern = NULL;
+		pgbrush->pattern = NULL;
+	}
 
-	if (pgbrush->blend)
+	if (pgbrush->blend) {
 		if (pgbrush->blend->count > 0) {
 			GdipFree (pgbrush->blend->factors);
 			GdipFree (pgbrush->blend->positions);
 		}
-	GdipFree (pgbrush->blend);
-	pgbrush->blend = NULL;
+		GdipFree (pgbrush->blend);
+		pgbrush->blend = NULL;
+	}
 	
-	if (pgbrush->presetColors)
+	if (pgbrush->presetColors) {
 		if (pgbrush->presetColors->count > 0) {
 			GdipFree (pgbrush->presetColors->colors);
 			GdipFree (pgbrush->presetColors->positions);
 		}
-	GdipFree (pgbrush->presetColors);
-	pgbrush->presetColors = NULL;
+		GdipFree (pgbrush->presetColors);
+		pgbrush->presetColors = NULL;
+	}
 
 	GdipFree (pgbrush);
 	return Ok;
 }
 
+static void
+add_color_stops_from_blend (cairo_pattern_t *pattern, Blend *blend, ARGB color1, ARGB color2)
+{
+	int index;
+	double sr, sg, sb, sa;
+	double er, eg, eb, ea;
+	double offset, factor;
+
+	sa = (color1 >> 24) & 0xFF;
+	sr = (color1 >> 16) & 0xFF;
+	sg = (color1 >> 8) & 0xFF;
+	sb = color1 & 0xFF;
+
+	ea = (color2 >> 24) & 0xFF;
+	er = (color2 >> 16) & 0xFF;
+	eg = (color2 >> 8) & 0xFF;
+	eb = color2 & 0xFF;
+
+	for (index = 0; index < blend->count; index++) {
+		factor = blend->factors [index];
+		offset = blend->positions [index];
+
+		cairo_pattern_add_color_stop_rgba (pattern, offset, 
+					      ((sr * (1 - factor)) + (er * factor)) / 255,
+					      ((sg * (1 - factor)) + (eg * factor)) / 255,
+					      ((sb * (1 - factor)) + (eb * factor)) / 255,
+					      ((sa * (1 - factor)) + (ea * factor)) / 255);
+	}
+}
+
+static void
+add_color_stops_from_interpolation_colors (cairo_pattern_t *pattern, InterpolationColors *presetColors)
+{
+	int index;
+	double r, g, b, a;
+	double offset;
+	ARGB color;
+
+	/* MS accecpts positions above 1.0 also. Cairo assumes the values above 1.0 as 1.0
+	 * and values below 0 are assumed as 0. So we get different results if any of the
+	 * offset values is out of [0.0, 1.0].
+	 */
+	for (index = 0; index < presetColors->count; index++) {
+		color = presetColors->colors [index];
+		a = (color >> 24) & 0xFF;
+		r = (color >> 16) & 0xFF;
+		g = (color >> 8) & 0xFF;
+		b = color & 0xFF;
+		offset = presetColors->positions [index];
+
+		cairo_pattern_add_color_stop_rgba (pattern, offset, r / 255, g / 255, b / 255, a / 255);
+	}
+}
+
 GpStatus
 gdip_pgrad_setup (GpGraphics *graphics, GpBrush *brush)
 {
 	GpPathGradient *pgbrush;
+	GpStatus status;
 
-	g_return_val_if_fail (graphics != NULL, InvalidParameter);
-	g_return_val_if_fail (brush != NULL, InvalidParameter);
+	if (!graphics || !brush)
+		return InvalidParameter;
 
 	pgbrush = (GpPathGradient *) brush;
 
@@ -238,24 +300,57 @@ gdip_pgrad_setup (GpGraphics *graphics, GpBrush *brush)
 	/* We create the new pattern for brush, if the brush is changed
 	 * or if pattern has not been created yet.
 	 */
-	if (pgbrush->base.changed || (pgbrush->pattern) == NULL) {
+	if (pgbrush->base.changed || !pgbrush->pattern) {
+		cairo_pattern_t *pat;
+		float r = MIN (pgbrush->rectangle.Width / 2, pgbrush->rectangle.Height / 2);
+		int i;
 
 		/* destroy the existing pattern */
-		if (pgbrush->pattern)
+		if (pgbrush->pattern) {
 			cairo_pattern_destroy (pgbrush->pattern);
+			pgbrush->pattern = NULL;
+		}
 
-		/* FIXME: To implement this function we need cairo to support
-		 * path gradients.
-		 * As a stub, we simply set the color to the center color.
+		/* FIXME: To fully implement this function we need cairo to support path gradients.
+		 * Right now we have radial gradient which can be used, in some cases, to get the right effect.
 		 */
-		cairo_set_source_rgba (graphics->ct,
-				      ARGB_RED_N(pgbrush->centerColor),
-				      ARGB_GREEN_N(pgbrush->centerColor),
-				      ARGB_BLUE_N(pgbrush->centerColor),
-				      ARGB_ALPHA_N(pgbrush->centerColor));
+
+		pat = cairo_pattern_create_radial (pgbrush->center.X, pgbrush->center.Y, 0.0f,
+			pgbrush->center.X, pgbrush->center.Y, r);
+		status = gdip_get_pattern_status (pat);
+		if (status != Ok)
+			return status;
+
+		cairo_pattern_set_matrix (pat, &pgbrush->transform);
+
+		if ((pgbrush->blend->count > 1) && (pgbrush->boundaryColorsCount > 0)) {
+			/* FIXME: blending done using the a radial shape (not the path shape) */
+			add_color_stops_from_blend (pat, pgbrush->blend, pgbrush->boundaryColors[0], pgbrush->centerColor);
+		} else if (pgbrush->presetColors->count > 1) {
+			/* FIXME: copied from lineargradiantbrush, most probably not right */
+			add_color_stops_from_interpolation_colors (pat, pgbrush->presetColors);
+		} else {
+			cairo_pattern_add_color_stop_rgba (pat, 0.0f,
+				ARGB_RED_N (pgbrush->centerColor),
+				ARGB_GREEN_N (pgbrush->centerColor),
+				ARGB_BLUE_N (pgbrush->centerColor),
+				ARGB_ALPHA_N (pgbrush->centerColor));
+
+			/* if a single other boundary color is present, then we can do the a real radial */
+			if (pgbrush->boundaryColorsCount == 1) {
+				ARGB c = pgbrush->boundaryColors[0];
+				cairo_pattern_add_color_stop_rgba (pat, 1.0f,
+					ARGB_RED_N (c), ARGB_GREEN_N (c), ARGB_BLUE_N (c), ARGB_ALPHA_N (c));
+			} else {
+				/* FIXME: otherwise we (solid-)fill with the centerColor */
+			}
+		}
+
+		pgbrush->pattern = pat;
 	}
 
-	return Ok;
+	cairo_set_source (graphics->ct, pgbrush->pattern);
+	return gdip_get_status (cairo_status (graphics->ct));
 }
 
 GpPointF
@@ -276,35 +371,60 @@ gdip_get_center (GDIPCONST GpPointF *points, int count)
 	return center;
 }
 
+static BOOL
+gdip_has_non_empty_color (GDIPCONST ARGB *colors, int count)
+{
+	int i;
+	ARGB *c = (ARGB*)colors;
+
+	for (i = 0; i < count; i++, c++) {
+		if (*c)
+			return TRUE;
+	}
+
+	return FALSE;
+}
+
+/* coverity[+alloc : arg-*3] */
 GpStatus
 GdipCreatePathGradient (GDIPCONST GpPointF *points, int count, GpWrapMode wrapMode, GpPathGradient **polyGradient)
 {
 	int i;
-	GpPathData pdata;
 	GpPathGradient *gp;
 	GpPath *gppath = NULL;
+	GpStatus status;
+	GpPointF point;
 
-	g_return_val_if_fail (polyGradient != NULL, InvalidParameter);
-	g_return_val_if_fail (count >= 2, InvalidParameter);
+	if (!polyGradient)
+		return InvalidParameter;
+
+	/* this match MS GDI+ behaviour */
+	if (!points || (count < 2))
+		return OutOfMemory;
+
+	status = GdipCreatePath (FillModeAlternate, /* ??? */ &gppath);
+	if (status != Ok) {
+		if (gppath)
+			GdipDeletePath (gppath);
+		return status;
+	}
 
-	gp = gdip_pathgradient_new ();
-	
-	GdipCreatePath (FillModeAlternate, /* ??? */
-			&gppath);
 	GdipAddPathLine2 (gppath, points, count);
 
+	gp = gdip_pathgradient_new ();
 	gp->boundary = gppath;
 	gp->wrapMode = wrapMode;
 	gp->center = gdip_get_center (points, count);
 	gp->centerColor = MAKE_ARGB_ARGB(255,0,0,0); /* black center color */
     
 	/* set the bounding rectangle */
-	GdipGetPathData (gppath, &pdata);
+	point = g_array_index (gppath->points, GpPointF, 0);
 	/* set the first point as the edge of the rectangle */
-	gp->rectangle->X = pdata.Points[0].X;
-	gp->rectangle->Y = pdata.Points[0].Y;
-	for (i = 1; i < pdata.Count; i++) {
-		gdip_rect_expand_by (gp->rectangle, &(pdata.Points[i]));
+	gp->rectangle.X = point.X;
+	gp->rectangle.Y = point.Y;
+	for (i = 1; i < gppath->count; i++) {
+		point = g_array_index (gppath->points, GpPointF, i);
+		gdip_rect_expand_by (&gp->rectangle, &point);
 	}
 
 	*polyGradient = gp;
@@ -312,12 +432,25 @@ GdipCreatePathGradient (GDIPCONST GpPointF *points, int count, GpWrapMode wrapMo
 	return Ok;
 }
 
+/* coverity[+alloc : arg-*3] */
 GpStatus
 GdipCreatePathGradientI (GDIPCONST GpPoint *points, int count, GpWrapMode wrapMode, GpPathGradient **polyGradient)
 {
 	int i;
 	GpStatus result;
-	GpPointF *newPoints = GdipAlloc (sizeof (GpPointF) * count);
+	GpPointF *newPoints;
+
+	if (!polyGradient)
+		return InvalidParameter;
+
+	/* this match MS GDI+ behaviour */
+	if (!points || (count < 2))
+		return OutOfMemory;
+
+	newPoints = GdipAlloc (sizeof (GpPointF) * count);
+	if (!newPoints)
+		return OutOfMemory;
+
 	for (i = 0; i < count; i++) {
 		newPoints[i].X = points[i].X;
 		newPoints[i].Y = points[i].Y;
@@ -328,16 +461,19 @@ GdipCreatePathGradientI (GDIPCONST GpPoint *points, int count, GpWrapMode wrapMo
 	return result;
 }
 
+/* coverity[+alloc : arg-*1] */
 GpStatus
 GdipCreatePathGradientFromPath (GDIPCONST GpPath *path, GpPathGradient **polyGradient)
 {
 	int i, count;
 	GpPathGradient *gp;
 	GpPointF *points;
-	GpPathData pdata;
 
-	g_return_val_if_fail (path != NULL, InvalidParameter);
-	g_return_val_if_fail (polyGradient != NULL, InvalidParameter);
+	/* this match MS GDI+ behaviour */
+	if (!path || (path->count < 2))
+		return OutOfMemory;
+	if (!polyGradient)
+		return InvalidParameter;
 
 	gp = gdip_pathgradient_new ();
 	GdipClonePath ((GpPath *) path, &(gp->boundary));
@@ -348,12 +484,11 @@ GdipCreatePathGradientFromPath (GDIPCONST GpPath *path, GpPathGradient **polyGra
 	gp->centerColor = MAKE_ARGB_ARGB(255,255,255,255); /* white center color */
 
 	/* set the bounding rectangle */
-	GdipGetPathData (path, &pdata);
 	/* set the first point as the edge of the rectangle */
-	gp->rectangle->X = pdata.Points[0].X;
-	gp->rectangle->Y = pdata.Points[0].Y;
-	for (i = 1; i < pdata.Count; i++) {
-		gdip_rect_expand_by (gp->rectangle, &(pdata.Points[i]));
+	gp->rectangle.X = points [0].X;
+	gp->rectangle.Y = points [0].Y;
+	for (i = 1; i < count; i++) {
+		gdip_rect_expand_by (&gp->rectangle, &points[i]);
 	}
 
 	*polyGradient = gp;
@@ -365,8 +500,8 @@ GdipCreatePathGradientFromPath (GDIPCONST GpPath *path, GpPathGradient **polyGra
 GpStatus
 GdipGetPathGradientCenterColor (GpPathGradient *brush, ARGB *colors)
 {
-	g_return_val_if_fail (brush != NULL, InvalidParameter);
-	g_return_val_if_fail (colors != NULL, InvalidParameter);
+	if (!brush || !colors)
+		return InvalidParameter;
 
 	*colors = brush->centerColor;
 	return Ok;
@@ -375,7 +510,8 @@ GdipGetPathGradientCenterColor (GpPathGradient *brush, ARGB *colors)
 GpStatus
 GdipSetPathGradientCenterColor (GpPathGradient *brush, ARGB colors)
 {
-	g_return_val_if_fail (brush != NULL, InvalidParameter);
+	if (!brush)
+		return InvalidParameter;
 
 	brush->centerColor = colors;
 	brush->base.changed = TRUE;
@@ -386,9 +522,9 @@ GpStatus
 GdipGetPathGradientSurroundColorsWithCount (GpPathGradient *brush, ARGB *colors, int *count)
 {
 	int i;
-	g_return_val_if_fail (brush != NULL, InvalidParameter);
-	g_return_val_if_fail (colors != NULL, InvalidParameter);
-	g_return_val_if_fail (count != NULL, InvalidParameter);
+
+	if (!brush || !colors || !count)
+		return InvalidParameter;
 
 	for (i = 0; i < *count && i < brush->boundaryColorsCount; i++) {
 		colors[i] = brush->boundaryColors[i];
@@ -401,13 +537,16 @@ GdipGetPathGradientSurroundColorsWithCount (GpPathGradient *brush, ARGB *colors,
 GpStatus
 GdipSetPathGradientSurroundColorsWithCount (GpPathGradient *brush, GDIPCONST ARGB *colors, int *count)
 {
-	g_return_val_if_fail (brush != NULL, InvalidParameter);
-	g_return_val_if_fail (colors != NULL, InvalidParameter);
-	g_return_val_if_fail (count != NULL, InvalidParameter);
+	if (!brush || !colors || !count)
+		return InvalidParameter;
 
 	if ((*count < 1) || (*count > brush->boundary->count))
 		return InvalidParameter;
 
+	/* at least one of the colors must be non-zero (i.e. not empty) */
+	if (!gdip_has_non_empty_color (colors, *count))
+		return Ok;
+
 	if (*count != brush->boundaryColorsCount) {
 		GdipFree (brush->boundaryColors);
 		brush->boundaryColors = (ARGB *) GdipAlloc (sizeof(ARGB) * (*count));
@@ -421,8 +560,8 @@ GdipSetPathGradientSurroundColorsWithCount (GpPathGradient *brush, GDIPCONST ARG
 GpStatus
 GdipGetPathGradientCenterPoint (GpPathGradient *brush, GpPointF *point)
 {
-	g_return_val_if_fail (brush != NULL, InvalidParameter);
-	g_return_val_if_fail (point != NULL, InvalidParameter);
+	if (!brush || !point)
+		return InvalidParameter;
 
 	point->X = brush->center.X;
 	point->Y = brush->center.Y;
@@ -432,8 +571,8 @@ GdipGetPathGradientCenterPoint (GpPathGradient *brush, GpPointF *point)
 GpStatus
 GdipSetPathGradientCenterPoint (GpPathGradient *brush, GDIPCONST GpPointF *point)
 {
-	g_return_val_if_fail (brush != NULL, InvalidParameter);
-	g_return_val_if_fail (point != NULL, InvalidParameter);
+	if (!brush || !point)
+		return InvalidParameter;
 	
 	brush->center.X = point->X;
 	brush->center.Y = point->Y;
@@ -444,18 +583,18 @@ GdipSetPathGradientCenterPoint (GpPathGradient *brush, GDIPCONST GpPointF *point
 GpStatus
 GdipGetPathGradientRect (GpPathGradient *brush, GpRectF *rect)
 {
-	g_return_val_if_fail (brush != NULL, InvalidParameter);
-	g_return_val_if_fail (rect != NULL, InvalidParameter);
+	if (!brush || !rect)
+		return InvalidParameter;
 	
-	*rect = *(brush->rectangle);
+	memcpy (rect, &brush->rectangle, sizeof (GpRectF));
 	return Ok;
 }
 
 GpStatus
 GdipGetPathGradientSurroundColorCount (GpPathGradient *brush, int *count)
 {
-	g_return_val_if_fail (brush != NULL, InvalidParameter);
-	g_return_val_if_fail (count != NULL, InvalidParameter);
+	if (!brush || !count)
+		return InvalidParameter;
 	
 	*count = brush->boundaryColorsCount;
 	return Ok;
@@ -464,7 +603,8 @@ GdipGetPathGradientSurroundColorCount (GpPathGradient *brush, int *count)
 GpStatus
 GdipGetPathGradientBlendCount (GpPathGradient *brush, int *count)
 {
-	g_return_val_if_fail (brush != NULL, InvalidParameter);
+	if (!brush || !count)
+		return InvalidParameter;
 	
 	/* If count is less than 1, we are not in a proper state 
 	 * to return blend property. By default, we have one blend
@@ -481,10 +621,8 @@ GdipGetPathGradientBlendCount (GpPathGradient *brush, int *count)
 GpStatus
 GdipGetPathGradientBlend (GpPathGradient *brush, float *blend, float *positions, int count)
 {
-	g_return_val_if_fail (brush != NULL, InvalidParameter);
-	g_return_val_if_fail (blend != NULL, InvalidParameter);
-	g_return_val_if_fail (positions != NULL, InvalidParameter);
-	g_return_val_if_fail (brush->blend->count == count, InvalidParameter);
+	if (!brush || !blend || !positions || (brush->blend->count != count))
+		return InvalidParameter;
 
 	/* If count is less than 1, we are not in a proper state 
 	 * to return blend property. By default, we have one blend
@@ -506,18 +644,19 @@ GdipSetPathGradientBlend (GpPathGradient *brush, GDIPCONST float *blend, GDIPCON
 	float *blendPositions;
 	int index;
 
-	g_return_val_if_fail (brush != NULL, InvalidParameter);
-	g_return_val_if_fail (blend != NULL, InvalidParameter);
-	g_return_val_if_fail (positions != NULL, InvalidParameter);
-
-	g_return_val_if_fail (count >= 2, InvalidParameter);
+	if (!brush || !blend || !positions || (count < 2))
+		return InvalidParameter;
 
 	if (brush->blend->count != count) {
 		blendFactors = (float *) GdipAlloc (count * sizeof (float));
-		g_return_val_if_fail (blendFactors != NULL, OutOfMemory);
+		if (!blendFactors)
+			return OutOfMemory;
 
 		blendPositions = (float *) GdipAlloc (count * sizeof (float));
-		g_return_val_if_fail (blendPositions != NULL, OutOfMemory);
+		if (!blendPositions) {
+			GdipFree (blendFactors);
+			return OutOfMemory;
+		}
 
 		/* free the existing values */
 		if (brush->blend->count != 0) {
@@ -554,7 +693,8 @@ GdipSetPathGradientBlend (GpPathGradient *brush, GDIPCONST float *blend, GDIPCON
 GpStatus
 GdipGetPathGradientPresetBlendCount (GpPathGradient *brush, int *count)
 {
-	g_return_val_if_fail (brush != NULL, InvalidParameter);
+	if (!brush)
+		return InvalidParameter;
 
 	*count = brush->presetColors->count;
 	return Ok;
@@ -563,11 +703,9 @@ GdipGetPathGradientPresetBlendCount (GpPathGradient *brush, int *count)
 GpStatus
 GdipGetPathGradientPresetBlend (GpPathGradient *brush, ARGB *blend, float *positions, int count)
 {
-	g_return_val_if_fail (brush != NULL, InvalidParameter);
-	g_return_val_if_fail (blend != NULL, InvalidParameter);
-	g_return_val_if_fail (positions != NULL, InvalidParameter);
-	g_return_val_if_fail (brush->presetColors->count == count, InvalidParameter);
-	
+	if (!brush || !blend || !positions || (brush->presetColors->count != count) || (count < 2))
+		return InvalidParameter;
+
 	memcpy (blend, brush->presetColors->colors, count * sizeof (ARGB));
 	memcpy (positions, brush->presetColors->positions, count * sizeof (float));
 
@@ -581,17 +719,19 @@ GdipSetPathGradientPresetBlend (GpPathGradient *brush, GDIPCONST ARGB *blend, GD
 	float *blendPositions;
 	int index;
 
-	g_return_val_if_fail (brush != NULL, InvalidParameter);
-	g_return_val_if_fail (blend != NULL, InvalidParameter);
-	g_return_val_if_fail (positions != NULL, InvalidParameter);
-	g_return_val_if_fail (count >= 2, InvalidParameter);
+	if (!brush || !blend || !positions || (count < 2))
+		return InvalidParameter;
 
 	if (brush->presetColors->count != count) {
 		blendColors = (ARGB *) GdipAlloc (count * sizeof (ARGB));
-		g_return_val_if_fail (blendColors != NULL, OutOfMemory);
+		if (!blendColors)
+			return OutOfMemory;
 
 		blendPositions = (float *) GdipAlloc (count * sizeof (float));
-		g_return_val_if_fail (blendPositions != NULL, OutOfMemory);
+		if (!blendPositions) {
+			GdipFree (blendColors);
+			return OutOfMemory;
+		}
 
 		/* free the existing values */
 		if (brush->presetColors->count != 0) {
@@ -645,7 +785,8 @@ GdipSetPathGradientSigmaBlend (GpPathGradient *brush, float focus, float scale)
 	float curve_top;
 	float curve_height;
 
-	g_return_val_if_fail (brush != NULL, InvalidParameter);
+	if (!brush)
+		return InvalidParameter;
 
 	if (focus == 0 || focus == 1) {
 		count = 256;
@@ -653,10 +794,14 @@ GdipSetPathGradientSigmaBlend (GpPathGradient *brush, float focus, float scale)
 
 	if (brush->blend->count != count) {
 		blends = (float *) GdipAlloc (count * sizeof (float));
-		g_return_val_if_fail (blends != NULL, OutOfMemory);
+		if (!blends)
+			return OutOfMemory;
 
 		positions = (float *) GdipAlloc (count * sizeof (float));
-		g_return_val_if_fail (positions != NULL, OutOfMemory);
+		if (!positions) {
+			GdipFree (blends);
+			return OutOfMemory;
+		}
 
 		/* free the existing values */
 		if (brush->blend->count != 0) {
@@ -812,7 +957,8 @@ GdipSetPathGradientLinearBlend (GpPathGradient *brush, float focus, float scale)
 	float *positions;
 	int count = 3;
 
-	g_return_val_if_fail (brush != NULL, InvalidParameter);
+	if (!brush)
+		return InvalidParameter;
 
 	if (focus == 0 || focus == 1) {
 		count = 2;
@@ -820,10 +966,14 @@ GdipSetPathGradientLinearBlend (GpPathGradient *brush, float focus, float scale)
 
 	if (brush->blend->count != count) {
 		blends = (float *) GdipAlloc (count * sizeof (float));
-		g_return_val_if_fail (blends != NULL, OutOfMemory);
+		if (!blends)
+			return OutOfMemory;
 
 		positions = (float *) GdipAlloc (count * sizeof (float));
-		g_return_val_if_fail (positions != NULL, OutOfMemory);
+		if (!positions) {
+			GdipFree (blends);
+			return OutOfMemory;
+		}
 
 		/* free the existing values */
 		if (brush->blend->count != 0) {
@@ -879,8 +1029,8 @@ GdipSetPathGradientLinearBlend (GpPathGradient *brush, float focus, float scale)
 GpStatus
 GdipGetPathGradientWrapMode (GpPathGradient *brush, GpWrapMode *wrapMode)
 {
-	g_return_val_if_fail (brush != NULL, InvalidParameter);
-	g_return_val_if_fail (wrapMode != NULL, InvalidParameter);
+	if (!brush || !wrapMode)
+		return InvalidParameter;
 
 	*wrapMode = brush->wrapMode;
 	return Ok;
@@ -889,7 +1039,8 @@ GdipGetPathGradientWrapMode (GpPathGradient *brush, GpWrapMode *wrapMode)
 GpStatus
 GdipSetPathGradientWrapMode (GpPathGradient *brush, GpWrapMode wrapMode)
 {
-	g_return_val_if_fail (brush != NULL, InvalidParameter);
+	if (!brush)
+		return InvalidParameter;
 
 	brush->wrapMode = wrapMode;
 	brush->base.changed = TRUE;
@@ -899,8 +1050,8 @@ GdipSetPathGradientWrapMode (GpPathGradient *brush, GpWrapMode wrapMode)
 GpStatus
 GdipGetPathGradientTransform (GpPathGradient *brush, GpMatrix *matrix)
 {
-	g_return_val_if_fail (brush != NULL, InvalidParameter);
-	g_return_val_if_fail (matrix != NULL, InvalidParameter);
+	if (!brush || !matrix)
+		return InvalidParameter;
 
 	/* If presetcolors are set, we are not in a proper state 
 	 * to return transform property.
@@ -908,17 +1059,25 @@ GdipGetPathGradientTransform (GpPathGradient *brush, GpMatrix *matrix)
 	if (brush->presetColors->count >= 2)
 		return WrongState;
 
-	gdip_cairo_matrix_copy(matrix, &brush->transform);
+	gdip_cairo_matrix_copy (matrix, &brush->transform);
 	return Ok;
 }
 
 GpStatus
 GdipSetPathGradientTransform (GpPathGradient *brush, GpMatrix *matrix)
 {
-	g_return_val_if_fail (brush != NULL, InvalidParameter);
-	g_return_val_if_fail (matrix != NULL, InvalidParameter);
+	GpStatus status;
+	BOOL invertible;
 
-	gdip_cairo_matrix_copy(&brush->transform, matrix);
+	if (!brush || !matrix)
+		return InvalidParameter;
+
+	/* the matrix MUST be invertible to be used */
+	status = GdipIsMatrixInvertible ((GpMatrix*) matrix, &invertible);
+	if (!invertible || (status != Ok))
+		return InvalidParameter;
+
+	gdip_cairo_matrix_copy (&brush->transform, matrix);
 	brush->base.changed = TRUE;
 	return Ok;
 }
@@ -926,113 +1085,83 @@ GdipSetPathGradientTransform (GpPathGradient *brush, GpMatrix *matrix)
 GpStatus
 GdipResetPathGradientTransform (GpPathGradient *brush)
 {
-    g_return_val_if_fail (brush != NULL, InvalidParameter);
-    cairo_matrix_init_identity (brush->transform);
-    return Ok;
+	if (!brush)
+		return InvalidParameter;
+
+	cairo_matrix_init_identity (&brush->transform);
+	brush->base.changed = TRUE;
+	return Ok;
 }
 
 GpStatus
 GdipMultiplyPathGradientTransform (GpPathGradient *brush, GDIPCONST GpMatrix *matrix, GpMatrixOrder order)
 {
-	cairo_matrix_t *mat = NULL;
-	g_return_val_if_fail (brush != NULL, InvalidParameter);
-	g_return_val_if_fail (matrix != NULL, InvalidParameter);
+	GpStatus status;
+	BOOL invertible;
+	cairo_matrix_t mat;
 
-	GdipCreateMatrix (&mat);
-	
-	if (order == MatrixOrderPrepend)
-		cairo_matrix_multiply (mat, matrix, brush->transform);
-	else if (order == MatrixOrderAppend)
-		cairo_matrix_multiply (mat, brush->transform, matrix);
-	else {
+	if (!brush || !matrix)
 		return InvalidParameter;
-	}
 
-	gdip_cairo_matrix_copy (brush->transform, mat);
+	/* the matrix MUST be invertible to be used */
+	status = GdipIsMatrixInvertible ((GpMatrix*) matrix, &invertible);
+	if (!invertible || (status != Ok))
+		return InvalidParameter;
+
+	if (order == MatrixOrderPrepend)
+		cairo_matrix_multiply (&mat, matrix, &brush->transform);
+	else if (order == MatrixOrderAppend)
+		cairo_matrix_multiply (&mat, &brush->transform, matrix);
+
+	gdip_cairo_matrix_copy (&brush->transform, &mat);
+	brush->base.changed = TRUE;
 	return Ok;
 }
 
 GpStatus
 GdipTranslatePathGradientTransform (GpPathGradient *brush, float dx, float dy, GpMatrixOrder order)
 {
-	g_return_val_if_fail (brush != NULL, InvalidParameter);
+	GpStatus status;
 
-	if (order == MatrixOrderAppend) {
-		cairo_matrix_translate (brush->transform, dx, dy);
-	} else if (order == MatrixOrderPrepend) {
-		cairo_matrix_t *mat, *matres;
-		
-		GdipCreateMatrix (&mat);
-		GdipCreateMatrix (&matres);
-
-		cairo_matrix_init_identity (mat);
-		cairo_matrix_translate (mat, dx, dy);
-		cairo_matrix_multiply (matres, mat, brush->transform);
-		gdip_cairo_matrix_copy (brush->transform, matres);
-
-	} else {
+	if (!brush)
 		return InvalidParameter;
-	}
 
-	return Ok;
+	if ((status = GdipTranslateMatrix (&brush->transform, dx, dy, order)) == Ok)
+		brush->base.changed = TRUE;
+	return status;
 }
 
 GpStatus
 GdipScalePathGradientTransform (GpPathGradient *brush, float sx, float sy, GpMatrixOrder order)
 {
-	g_return_val_if_fail (brush != NULL, InvalidParameter);
+	GpStatus status;
 
-	if (order == MatrixOrderAppend) {
-		cairo_matrix_scale (brush->transform, sx, sy);
-	} else if (order == MatrixOrderPrepend) {
-		cairo_matrix_t *mat, *matres;
-		
-		GdipCreateMatrix (&mat);
-		GdipCreateMatrix (&matres);
-		
-		cairo_matrix_init_identity (mat);
-		cairo_matrix_scale (mat, sx, sy);
-		cairo_matrix_multiply (matres, mat, brush->transform);
-		gdip_cairo_matrix_copy (brush->transform, matres);
-
-	} else {
+	if (!brush)
 		return InvalidParameter;
-	}
 
-	return Ok;
+	if ((status = GdipScaleMatrix (&brush->transform, sx, sy, order)) == Ok)
+		brush->base.changed = TRUE;
+	return status;
 }
 
 GpStatus
 GdipRotatePathGradientTransform (GpPathGradient *brush, float angle, GpMatrixOrder order)
 {
-	g_return_val_if_fail (brush != NULL, InvalidParameter);
+	GpStatus status;
 
-	if (order == MatrixOrderAppend) {
-		cairo_matrix_rotate (brush->transform, angle * DEGTORAD);
-	} else if (order == MatrixOrderPrepend) {
-		cairo_matrix_t *mat, *matres;
-		
-		GdipCreateMatrix (&mat);
-		GdipCreateMatrix (&matres);
-				
-		cairo_matrix_init_identity (mat);
-		cairo_matrix_rotate (mat, angle * DEGTORAD);
-		cairo_matrix_multiply (matres, mat, brush->transform);
-		gdip_cairo_matrix_copy (brush->transform, matres);
-
-	} else {
+	if (!brush)
 		return InvalidParameter;
-	}
 
-	return Ok;
+	if ((status = GdipRotateMatrix (&brush->transform, angle, order)) == Ok)
+		brush->base.changed = TRUE;
+	return status;
 }
 
 GpStatus
 GdipGetPathGradientFocusScales (GpPathGradient *brush, float *xScale, float *yScale)
 {
-	g_return_val_if_fail (brush != NULL, InvalidParameter);
-	g_return_val_if_fail (xScale != NULL, InvalidParameter);
-	g_return_val_if_fail (yScale != NULL, InvalidParameter);
+	if (!brush || !xScale || !yScale)
+		return InvalidParameter;
 
 	*xScale = brush->focusScales.X;
 	*yScale = brush->focusScales.Y;
@@ -1042,9 +1171,11 @@ GdipGetPathGradientFocusScales (GpPathGradient *brush, float *xScale, float *ySc
 GpStatus
 GdipSetPathGradientFocusScales (GpPathGradient *brush, float xScale, float yScale)
 {
-	g_return_val_if_fail (brush != NULL, InvalidParameter);
+	if (!brush)
+		return InvalidParameter;
 
 	brush->focusScales.X = xScale;
 	brush->focusScales.Y = yScale;
+	brush->base.changed = TRUE;
 	return Ok;
 }
