@@ -2,6 +2,7 @@
  * pen.c
  * 
  * Copyright (c) 2003 Alexandre Pigolkine
+ * Copyright (C) 2006 Novell, Inc (http://www.novell.com)
  * 
  * Permission is hereby granted, free of charge, to any person obtaining a copy of this software 
  * and associated documentation files (the "Software"), to deal in the Software without restriction, 
@@ -39,6 +40,8 @@ gdip_pen_init (GpPen *pen)
         pen->line_join = LineJoinMiter;
 	pen->dash_style = DashStyleSolid;
 	pen->line_cap = LineCapFlat;
+	pen->end_cap = LineCapFlat;		/* ignored, Cairo only support a single start/end line cap */
+	pen->dash_cap = DashCapFlat;		/* ignored */
 	pen->mode = PenAlignmentCenter;
 	pen->dash_offset = 0;
 	pen->dash_count = 0;
@@ -48,8 +51,7 @@ gdip_pen_init (GpPen *pen)
 	pen->compound_array = NULL;
 	pen->unit = UnitWorld;
 	pen->changed = TRUE;
-	GdipCreateMatrix (&pen->matrix);
-	cairo_matrix_init_identity (pen->matrix);
+	cairo_matrix_init_identity (&pen->matrix);
 }
 
 GpPen*
@@ -58,7 +60,7 @@ gdip_pen_new (void)
         GpPen *result = (GpPen *) GdipAlloc (sizeof (GpPen));
 
 	if (result)
-	  gdip_pen_init (result);
+		gdip_pen_init (result);
 
         return result;
 }
@@ -84,11 +86,20 @@ convert_line_join (GpLineJoin join)
 }
 
 static cairo_line_cap_t
-convert_line_cap (GpLineCap cap)
+convert_line_cap (GpPen *pen)
 {
-        switch (cap) {
-        
+        switch (pen->line_cap) {
+
+	/* HACK - this keeps SWF (mostly) happy with results very similar to GDI+
+	 * (under those specific cases) and also keeps the pen's functionalities 
+	 * on par with GDI+
+	 */
         case LineCapFlat:
+		if (pen->dash_array || (pen->width > 1.0))
+	                return CAIRO_LINE_CAP_BUTT;
+		else
+	                return CAIRO_LINE_CAP_SQUARE;
+        
         case LineCapSquare:
                 return CAIRO_LINE_CAP_SQUARE;
 
@@ -108,7 +119,7 @@ convert_line_cap (GpLineCap cap)
 }
 
 static double *
-convert_dash_array (float *f, float width, int count)
+convert_dash_array (float *f, double width, int count)
 {
         double *retval = GdipAlloc (sizeof (double) * count);
         int i;
@@ -123,17 +134,17 @@ GpStatus
 gdip_pen_setup (GpGraphics *graphics, GpPen *pen)
 {
 	GpStatus status;
-	GpMatrix *product = NULL;
-	
-	g_return_val_if_fail (graphics != NULL, InvalidParameter);
-	g_return_val_if_fail (pen != NULL, InvalidParameter);
+	cairo_matrix_t product;
+	double widthx;
+
+	if (!graphics || !pen)
+		return InvalidParameter;
 
 	status = gdip_brush_setup (graphics, pen->brush);
 	if (status != Ok)
 		return status;
 
-	GdipCreateMatrix (&product);
-	cairo_matrix_init_identity (product);
+	cairo_matrix_init_identity (&product);
 
 	/* Here we use product of pen->matrix and graphics->copy_of_ctm.
 	 * This gives us absolute results with respect to graphics. We
@@ -142,10 +153,8 @@ gdip_pen_setup (GpGraphics *graphics, GpPen *pen)
 	 * every time we perform stroke operations. Graphics matrix gets
 	 * reset to its own state after stroking.
 	 */
-	cairo_matrix_multiply (product, pen->matrix, graphics->copy_of_ctm);
-	cairo_set_matrix (graphics->ct, product);
-
-	GdipDeleteMatrix (product);
+	cairo_matrix_multiply (&product, &pen->matrix, graphics->copy_of_ctm);
+	cairo_set_matrix (graphics->ct, &product);
 
 	/* Don't need to setup, if pen is the same as the cached pen and
 	 * it is not changed. Just comparing pointers may not be sufficient
@@ -157,21 +166,25 @@ gdip_pen_setup (GpGraphics *graphics, GpPen *pen)
 		return Ok;
 
 	if (pen->width <= 0) { /* we draw a pixel wide line if width is <=0 */
-		double widthx = 1.0;
 		double widthy = 1.0;
+		widthx = 1.0;
+
 		cairo_device_to_user_distance (graphics->ct, &widthx, &widthy);
-		cairo_set_line_width (graphics->ct, widthx);
-	} else
-		cairo_set_line_width (graphics->ct, (double) pen->width);
+	} else {
+		widthx = (double) pen->width;
+	}
+	cairo_set_line_width (graphics->ct, widthx);
 
         cairo_set_miter_limit (graphics->ct, (double) pen->miter_limit);
         cairo_set_line_join (graphics->ct, convert_line_join (pen->line_join));
-        cairo_set_line_cap (graphics->ct, convert_line_cap (pen->line_cap));
+        cairo_set_line_cap (graphics->ct, convert_line_cap (pen));
 
         if (pen->dash_count > 0) {
                 double *dash_array;
 
-                dash_array = convert_dash_array (pen->dash_array, pen->width, pen->dash_count);
+		/* note: pen->width may be different from what was used to 
+		   call cairo_set_line_width, e.g. 0.0 (#78742) */
+                dash_array = convert_dash_array (pen->dash_array, widthx, pen->dash_count);
                 cairo_set_dash (graphics->ct, dash_array, pen->dash_count, pen->dash_offset);
                 GdipFree (dash_array);
         } else /* Clear the dashes, if set in previous calls */
@@ -184,66 +197,99 @@ gdip_pen_setup (GpGraphics *graphics, GpPen *pen)
 	return gdip_get_status (cairo_status (graphics->ct));
 }
 
+// coverity[+alloc : arg-*3]
 GpStatus 
 GdipCreatePen1 (int argb, float width, GpUnit unit, GpPen **pen)
 {
         GpStatus s;
-	GpSolidFill *solidBrush;
+	GpSolidFill *solidBrush = NULL;
+	GpPen *result;
 
-	g_return_val_if_fail (pen != NULL, InvalidParameter);
+	if (!pen)
+		return InvalidParameter;
 
-        *pen = gdip_pen_new ();
-	g_return_val_if_fail (*pen != NULL, InvalidParameter);
+	*pen = result = gdip_pen_new ();
+	if (!result)
+		return OutOfMemory;
 
-        (*pen)->color = argb;
-
-        /* FIXME: do unit conversion when setting width */
-        (*pen)->width = width;
+	result->color = argb;
+	/* FIXME: do unit conversion when setting width */
+	result->width = width;
 
 	s = GdipCreateSolidFill (argb, &solidBrush);
-        (*pen)->brush = (GpBrush*) solidBrush;
-	(*pen)->own_brush = TRUE;
-        return s;
+	if (s != Ok) {
+		if (solidBrush)
+			GdipDeleteBrush ((GpBrush*) solidBrush);
+		GdipFree (result);
+		*pen = NULL;
+		return s;
+	}
+
+	result->brush = (GpBrush*) solidBrush;
+	result->own_brush = TRUE;
+	return Ok;
 }
 
+// coverity[+alloc : arg-*3]
 GpStatus
 GdipCreatePen2 (GpBrush *brush, float width, GpUnit unit, GpPen **pen)
 {
         int color;
         GpStatus s;
         GpBrushType type;
+	GpPen *result;
 
-	g_return_val_if_fail (brush != NULL, InvalidParameter);
-	g_return_val_if_fail (pen != NULL, InvalidParameter);
+	if (!brush || !pen)
+		return InvalidParameter;
 
-        *pen = gdip_pen_new ();
-
-	g_return_val_if_fail (*pen != NULL, OutOfMemory);
+        *pen = result = gdip_pen_new ();
+	if (!result)
+		return OutOfMemory;
 
         /* FIXME: do unit conversion when setting width */
-        (*pen)->width = width;
-	(*pen)->brush = brush;
+        result->width = width;
+
+	/* the user supplied brush can be disposed, we must clone it to ensure
+	 * it's valid when we need to set the pen */
+	s = GdipCloneBrush (brush, &result->brush);
+	if (s != Ok) {
+		GdipFree (result);
+		*pen = NULL;
+		return s;
+	}
+
+	result->own_brush = TRUE;
 
         s = GdipGetBrushType (brush, &type);
-	if (s != Ok)
+	if (s != Ok) {
+		GdipFree (result);
+		*pen = NULL;
 		return s;
+	}
 
         switch (type) {
 
         case BrushTypeSolidColor:
                 s = GdipGetSolidFillColor ((GpSolidFill*) brush, &color);
-		if (s != Ok)
+		if (s != Ok) {
+			GdipFree (result);
+			*pen = NULL;
 			return s;
-                (*pen)->color = color;
+		}
+                result->color = color;
+		*pen = result;
                 return Ok;
 
         case BrushTypeHatchFill:
         case BrushTypeTextureFill:
         case BrushTypePathGradient:
         case BrushTypeLinearGradient:
+		*pen = result;
 		return Ok;
 
         default:
+		GdipFree (result);
+		*pen = NULL;
                 return GenericError;
         }
 }
@@ -259,16 +305,16 @@ clone_dash_array (float *clone, float *array, int size)
         return clone;
 }
 
+// coverity[+alloc : arg-*1]
 GpStatus 
 GdipClonePen (GpPen *pen, GpPen **clonepen)
 {
         GpPen *result;
-        GpMatrix *matrix = NULL;        /* copy of pen->matrix */
         float *dashes;                  /* copy off pen->dash_array */
         float *compound_array = NULL;   /* copy off pen->compound_array */
 
-	g_return_val_if_fail (pen != NULL, InvalidParameter);
-	g_return_val_if_fail (clonepen != NULL, InvalidParameter);
+	if (!pen || !clonepen)
+		return InvalidParameter;
 
 	/* we make a copy of dash array only if it is owned by pen, i.e. it is not
 	 * our global array.
@@ -276,7 +322,8 @@ GdipClonePen (GpPen *pen, GpPen **clonepen)
 
 	if (pen->dash_count > 0 && pen->own_dash_array) {
 		dashes = (float *) GdipAlloc (pen->dash_count * sizeof (float));
-		g_return_val_if_fail (dashes != NULL, OutOfMemory);
+		if (!dashes)
+			return OutOfMemory;
 		clone_dash_array (dashes, pen->dash_array, pen->dash_count);
 	} else
 		dashes = pen->dash_array;
@@ -291,19 +338,8 @@ GdipClonePen (GpPen *pen, GpPen **clonepen)
 		clone_dash_array (compound_array, pen->compound_array, pen->compound_count);
 	}
 
-	GdipCreateMatrix (&matrix);
-        if (GdipCloneMatrix (pen->matrix, &matrix) != Ok) {
-		if (pen->dash_count > 0)
-			GdipFree (dashes);
-		if (pen->compound_count > 0)
-			GdipFree (compound_array);
-
-		return OutOfMemory;
-	}
-
 	result = gdip_pen_new ();
 	if (result == NULL) {
-		GdipDeleteMatrix (matrix);
 		if (pen->dash_count > 0)
 			GdipFree (dashes);
 		if (pen->compound_count > 0)
@@ -328,6 +364,7 @@ GdipClonePen (GpPen *pen, GpPen **clonepen)
         result->line_join = pen->line_join;
 	result->dash_style = pen->dash_style;
         result->line_cap = pen->line_cap;
+	result->end_cap = pen->end_cap;
 	result->mode = pen->mode;
         result->dash_offset = pen->dash_offset;
 	result->dash_count = pen->dash_count;
@@ -336,9 +373,7 @@ GdipClonePen (GpPen *pen, GpPen **clonepen)
 	result->compound_count = pen->compound_count;
 	result->compound_array = compound_array;
 	result->unit = pen->unit;
-	if (result->matrix)
-		GdipDeleteMatrix (result->matrix);
-	result->matrix = matrix;
+	gdip_cairo_matrix_copy (&result->matrix, &pen->matrix);
 	result->changed = pen->changed;
 
         *clonepen = result;
@@ -349,7 +384,8 @@ GdipClonePen (GpPen *pen, GpPen **clonepen)
 GpStatus 
 GdipDeletePen (GpPen *pen)
 {
-	g_return_val_if_fail (pen != NULL, InvalidParameter);
+	if (!pen)
+		return InvalidParameter;
 
         if (pen->dash_count != 0 && pen->own_dash_array) {
                 GdipFree (pen->dash_array);
@@ -358,7 +394,7 @@ GdipDeletePen (GpPen *pen)
 	}
 
 	if (pen->own_brush && pen->brush) {
-		GdipFree (pen->brush);
+		GdipDeleteBrush (pen->brush);
 		pen->brush = NULL;
 	}
 
@@ -368,20 +404,15 @@ GdipDeletePen (GpPen *pen)
 		pen->compound_count = 0;
 	}
 
-	if (pen->matrix) {
-		GdipDeleteMatrix (pen->matrix);
-		pen->matrix = NULL;
-	}
-
         GdipFree (pen);
-
 	return Ok;
 }
 
 GpStatus
 GdipSetPenWidth (GpPen *pen, float width)
 {
-	g_return_val_if_fail (pen != NULL, InvalidParameter);
+	if (!pen)
+		return InvalidParameter;
 
         pen->width = width;
 	pen->changed = TRUE;
@@ -391,8 +422,8 @@ GdipSetPenWidth (GpPen *pen, float width)
 GpStatus
 GdipGetPenWidth (GpPen *pen, float *width)
 {
-	g_return_val_if_fail (pen != NULL, InvalidParameter);
-	g_return_val_if_fail (width != NULL, InvalidParameter);
+	if (!pen || !width)
+		return InvalidParameter;
 
         *width = pen->width;
         return Ok;
@@ -405,8 +436,8 @@ GdipSetPenBrushFill (GpPen *pen, GpBrush *brush)
         int color;
         GpBrushType type;
 
-	g_return_val_if_fail (pen != NULL, InvalidParameter);
-	g_return_val_if_fail (brush != NULL, InvalidParameter);
+	if (!pen || !brush)
+		return InvalidParameter;
 
         s = GdipGetBrushType (brush, &type);
 
@@ -423,7 +454,7 @@ GdipSetPenBrushFill (GpPen *pen, GpBrush *brush)
 		pen->color = 0;
 
 	if (pen->own_brush && pen->brush)
-		GdipFree (pen->brush);
+		GdipDeleteBrush (pen->brush);
 
         pen->brush = brush;
 	pen->changed = TRUE;
@@ -435,8 +466,8 @@ GdipSetPenBrushFill (GpPen *pen, GpBrush *brush)
 GpStatus
 GdipGetPenBrushFill (GpPen *pen, GpBrush **brush)
 {
-	g_return_val_if_fail (pen != NULL, InvalidParameter);
-	g_return_val_if_fail (brush != NULL, InvalidParameter);
+	if (!pen || !brush)
+		return InvalidParameter;
 
         *brush = pen->brush;
 
@@ -446,11 +477,11 @@ GdipGetPenBrushFill (GpPen *pen, GpBrush **brush)
 GpStatus
 GdipGetPenFillType (GpPen *pen, GpPenType *type)
 {
-	g_return_val_if_fail (pen != NULL, InvalidParameter);
-	g_return_val_if_fail (type != NULL, InvalidParameter);
+	if (!pen || !type)
+		return InvalidParameter;
 
 	if (pen->brush != NULL)
-	  return GdipGetBrushType (pen->brush, (GpBrushType *) type);
+		return GdipGetBrushType (pen->brush, (GpBrushType *) type);
 
 	*type = PenTypeSolidColor;
 
@@ -460,19 +491,19 @@ GdipGetPenFillType (GpPen *pen, GpPenType *type)
 GpStatus
 GdipSetPenColor (GpPen *pen, int argb)
 {
-	g_return_val_if_fail (pen != NULL, InvalidParameter);
+	if (!pen)
+		return InvalidParameter;
 
+	pen->changed = (pen->color != argb);
         pen->color = argb;
-	pen->changed = TRUE;
-
         return Ok;
 }
 
 GpStatus
 GdipGetPenColor (GpPen *pen, int *argb)
 {
-	g_return_val_if_fail (pen != NULL, InvalidParameter);
-	g_return_val_if_fail (argb != NULL, InvalidParameter);
+	if (!pen || !argb)
+		return InvalidParameter;
 
         *argb = pen->color;
         return Ok;
@@ -481,18 +512,22 @@ GdipGetPenColor (GpPen *pen, int *argb)
 GpStatus
 GdipSetPenMiterLimit (GpPen *pen, float miterLimit)
 {
-	g_return_val_if_fail (pen != NULL, InvalidParameter);
+	if (!pen)
+		return InvalidParameter;
 
+	if (miterLimit < 1.0f)
+		miterLimit = 1.0f;
+
+	pen->changed = (pen->miter_limit != miterLimit);
         pen->miter_limit = miterLimit;
-	pen->changed = TRUE;
         return Ok;
 }
 
 GpStatus
 GdipGetPenMiterLimit (GpPen *pen, float *miterLimit)
 {
-	g_return_val_if_fail (pen != NULL, InvalidParameter);
-	g_return_val_if_fail (miterLimit != NULL, InvalidParameter);
+	if (!pen || !miterLimit)
+		return InvalidParameter;
 
         *miterLimit = pen->miter_limit;
         return Ok;
@@ -501,18 +536,19 @@ GdipGetPenMiterLimit (GpPen *pen, float *miterLimit)
 GpStatus
 GdipSetPenLineJoin (GpPen *pen, GpLineJoin lineJoin)
 {
-	g_return_val_if_fail (pen != NULL, InvalidParameter);
+	if (!pen)
+		return InvalidParameter;
 
+	pen->changed = (pen->line_join != lineJoin);
         pen->line_join = lineJoin;
-	pen->changed = TRUE;
         return Ok;
 }
 
 GpStatus
 GdipGetPenLineJoin (GpPen *pen, GpLineJoin *lineJoin)
 {
-	g_return_val_if_fail (pen != NULL, InvalidParameter);
-	g_return_val_if_fail (lineJoin != NULL, InvalidParameter);
+	if (!pen || !lineJoin)
+		return InvalidParameter;
 
         *lineJoin = pen->line_join;
         return Ok;
@@ -521,14 +557,23 @@ GdipGetPenLineJoin (GpPen *pen, GpLineJoin *lineJoin)
 GpStatus
 GdipSetPenLineCap197819 (GpPen *pen, GpLineCap startCap, GpLineCap endCap, GpDashCap dashCap)
 {
-	g_return_val_if_fail (pen != NULL, InvalidParameter);
+	if (!pen)
+		return InvalidParameter;
 
 	/* FIXME:
 	 * Cairo supports only one cap for a line. We use startcap for that.
 	 * Use end cap and dash cap when Cairo supports different caps for the
 	 * line ends and dashcap.
 	 */
+
 	pen->line_cap = startCap;
+	pen->end_cap = endCap;
+	/* any invalid value is changed to DashCapFlat */
+	if ((dashCap == DashCapRound) || (dashCap == DashCapTriangle))
+		pen->dash_cap = dashCap;
+	else
+		pen->dash_cap = DashCapFlat; /* default */
+
 	pen->changed = TRUE;
         return Ok;
 }
@@ -536,18 +581,19 @@ GdipSetPenLineCap197819 (GpPen *pen, GpLineCap startCap, GpLineCap endCap, GpDas
 GpStatus
 GdipSetPenMode (GpPen *pen, GpPenAlignment penMode)
 {
-	g_return_val_if_fail (pen != NULL, InvalidParameter);
+	if (!pen)
+		return InvalidParameter;
 
-        pen->mode = penMode;
-	pen->changed = TRUE;
-        return Ok;
+	pen->changed = (penMode != pen->mode);
+	pen->mode = penMode;
+	return Ok;
 }
 
 GpStatus
 GdipGetPenMode (GpPen *pen, GpPenAlignment *penMode)
 {
-	g_return_val_if_fail (pen != NULL, InvalidParameter);
-	g_return_val_if_fail (penMode != NULL, InvalidParameter);
+	if (!pen || !penMode)
+		return InvalidParameter;
 
         *penMode = pen->mode;
         return Ok;
@@ -556,8 +602,8 @@ GdipGetPenMode (GpPen *pen, GpPenAlignment *penMode)
 GpStatus
 GdipGetPenUnit (GpPen *pen, GpUnit *unit)
 {
-	g_return_val_if_fail (pen != NULL, InvalidParameter);
-	g_return_val_if_fail (unit != NULL, InvalidParameter);
+	if (!pen || !unit)
+		return InvalidParameter;
 
         *unit = pen->unit;
         return Ok;
@@ -566,7 +612,8 @@ GdipGetPenUnit (GpPen *pen, GpUnit *unit)
 GpStatus
 GdipSetPenUnit (GpPen *pen, GpUnit unit)
 {
-	g_return_val_if_fail (pen != NULL, InvalidParameter);
+	if (!pen)
+		return InvalidParameter;
 
         pen->unit = unit;
 	pen->changed = TRUE;
@@ -576,11 +623,18 @@ GdipSetPenUnit (GpPen *pen, GpUnit unit)
 GpStatus
 GdipSetPenTransform (GpPen *pen, GDIPCONST GpMatrix *matrix)
 {
-	g_return_val_if_fail (pen != NULL, InvalidParameter);
-	g_return_val_if_fail (matrix != NULL, InvalidParameter);
+	GpStatus status;
+	BOOL invertible;
 
+	if (!pen || !matrix)
+		return InvalidParameter;
 
-	gdip_cairo_matrix_copy(&pen->matrix, matrix);
+	/* the matrix MUST be invertible to be used */
+	status = GdipIsMatrixInvertible ((GpMatrix*) matrix, &invertible);
+	if (!invertible || (status != Ok))
+		return InvalidParameter;
+
+	gdip_cairo_matrix_copy (&pen->matrix, matrix);
 	pen->changed = TRUE;
         return Ok;
 }
@@ -588,8 +642,8 @@ GdipSetPenTransform (GpPen *pen, GDIPCONST GpMatrix *matrix)
 GpStatus
 GdipGetPenTransform (GpPen *pen, GpMatrix *matrix)
 {
-	g_return_val_if_fail (pen != NULL, InvalidParameter);
-	g_return_val_if_fail (matrix != NULL, InvalidParameter);
+	if (!pen || !matrix)
+		return InvalidParameter;
 
 	gdip_cairo_matrix_copy(matrix, &pen->matrix);
         return Ok;
@@ -598,9 +652,10 @@ GdipGetPenTransform (GpPen *pen, GpMatrix *matrix)
 GpStatus
 GdipResetPenTransform (GpPen *pen)
 {
-	g_return_val_if_fail (pen != NULL, InvalidParameter);
+	if (!pen)
+		return InvalidParameter;
 
-	cairo_matrix_init_identity (pen->matrix);
+	cairo_matrix_init_identity (&pen->matrix);
 
 	pen->changed = TRUE;
 
@@ -611,11 +666,21 @@ GpStatus
 GdipMultiplyPenTransform (GpPen *pen, GpMatrix *matrix, GpMatrixOrder order)
 {
 	GpStatus status;
-	g_return_val_if_fail (pen != NULL, InvalidParameter);
-	g_return_val_if_fail (matrix != NULL, InvalidParameter);
+	BOOL invertible;
 
-	status = GdipMultiplyMatrix (pen->matrix, matrix, order);
+	if (!pen || !matrix)
+		return InvalidParameter;
 
+	/* the matrix MUST be invertible to be used */
+	status = GdipIsMatrixInvertible ((GpMatrix*) matrix, &invertible);
+	if (!invertible || (status != Ok))
+		return InvalidParameter;
+
+	/* invalid GpMatrixOrder values are computed as MatrixOrderAppend (i.e. no error) */
+	if (order != MatrixOrderPrepend)
+		order = MatrixOrderAppend;
+
+	status = GdipMultiplyMatrix (&pen->matrix, matrix, order);
 	if (status == Ok)
 		pen->changed = TRUE;
 
@@ -626,9 +691,11 @@ GpStatus
 GdipTranslatePenTransform (GpPen *pen, float dx, float dy, GpMatrixOrder order)
 {
 	GpStatus status;
-	g_return_val_if_fail (pen != NULL, InvalidParameter);
 
-	status = GdipTranslateMatrix (pen->matrix, dx, dy, order);
+	if (!pen)
+		return InvalidParameter;
+
+	status = GdipTranslateMatrix (&pen->matrix, dx, dy, order);
 
 	if (status == Ok)
 		pen->changed = TRUE;
@@ -640,9 +707,11 @@ GpStatus
 GdipScalePenTransform (GpPen *pen, float sx, float sy, GpMatrixOrder order)
 {
 	GpStatus status;
-	g_return_val_if_fail (pen != NULL, InvalidParameter);
 
-        status = GdipScaleMatrix (pen->matrix, sx, sy, order);
+	if (!pen)
+		return InvalidParameter;
+
+        status = GdipScaleMatrix (&pen->matrix, sx, sy, order);
 
 	if (status == Ok)
 		pen->changed = TRUE;
@@ -654,9 +723,11 @@ GpStatus
 GdipRotatePenTransform (GpPen *pen, float angle, GpMatrixOrder order)
 {
 	GpStatus status;
-	g_return_val_if_fail (pen != NULL, InvalidParameter);
 
-	status = GdipRotateMatrix (pen->matrix, angle, order);
+	if (!pen)
+		return InvalidParameter;
+
+	status = GdipRotateMatrix (&pen->matrix, angle, order);
 
 	if (status == Ok)
 		pen->changed = TRUE;
@@ -667,15 +738,14 @@ GdipRotatePenTransform (GpPen *pen, float angle, GpMatrixOrder order)
 GpStatus
 GdipGetPenDashStyle (GpPen *pen, GpDashStyle *dashStyle)
 {
-	g_return_val_if_fail (pen != NULL, InvalidParameter);
-	g_return_val_if_fail (dashStyle != NULL, InvalidParameter);
+	if (!pen || !dashStyle)
+		return InvalidParameter;
 
         *dashStyle = pen->dash_style;
         return Ok;
 }
 
-static float Custom [] = { 1.0 };
-static float Dot []  = { 1.0, 2.0 };	// FIXME - this is a workaround for cairo, should be 1.0, 1.0
+static float Dot []  = { 1.0, 1.0 };
 static float Dash []  = { 3.0, 1.0 };
 static float DashDot [] = { 3.0, 1.0, 1.0, 1.0 };
 static float DashDotDot [] = { 3.0, 1.0, 1.0, 1.0, 1.0, 1.0 };
@@ -683,7 +753,8 @@ static float DashDotDot [] = { 3.0, 1.0, 1.0, 1.0, 1.0, 1.0 };
 GpStatus
 GdipSetPenDashStyle (GpPen *pen, GpDashStyle dashStyle)
 {
-	g_return_val_if_fail (pen != NULL, InvalidParameter);
+	if (!pen)
+		return InvalidParameter;
 
 	switch (dashStyle) {
 	case DashStyleSolid:
@@ -712,8 +783,8 @@ GdipSetPenDashStyle (GpPen *pen, GpDashStyle dashStyle)
 		break;
 
 	case DashStyleCustom:
-		pen->dash_array = Custom;
-		pen->dash_count = 1;
+		/* we keep the current assigned value when switching to Custom */
+		/* other special stuff happens in System.Drawing (but not here) */
 		break;
 
 	default:
@@ -728,8 +799,8 @@ GdipSetPenDashStyle (GpPen *pen, GpDashStyle dashStyle)
 GpStatus
 GdipGetPenDashOffset (GpPen *pen, float *offset)
 {
-	g_return_val_if_fail (pen != NULL, InvalidParameter);
-	g_return_val_if_fail (offset != NULL, InvalidParameter);
+	if (!pen || !offset)
+		return InvalidParameter;
 
         *offset = pen->dash_offset;
         return Ok;
@@ -738,7 +809,8 @@ GdipGetPenDashOffset (GpPen *pen, float *offset)
 GpStatus
 GdipSetPenDashOffset (GpPen *pen, float offset)
 {
-	g_return_val_if_fail (pen != NULL, InvalidParameter);
+	if (!pen)
+		return InvalidParameter;
 
 	pen->dash_offset = offset;
 	pen->changed = TRUE;
@@ -748,8 +820,8 @@ GdipSetPenDashOffset (GpPen *pen, float offset)
 GpStatus
 GdipGetPenDashCount (GpPen *pen, int *count)
 {
-	g_return_val_if_fail (pen != NULL, InvalidParameter);
-	g_return_val_if_fail (count != NULL, InvalidParameter);
+	if (!pen || !count)
+		return InvalidParameter;
 
         *count = pen->dash_count;
         return Ok;
@@ -761,9 +833,12 @@ GdipGetPenDashCount (GpPen *pen, int *count)
 GpStatus
 GdipGetPenDashArray (GpPen *pen, float *dash, int count)
 {
-	g_return_val_if_fail (pen != NULL, InvalidParameter);
-	g_return_val_if_fail (dash != NULL, InvalidParameter);
-	g_return_val_if_fail (count == pen->dash_count, InvalidParameter);
+	if (!pen || !dash)
+		return InvalidParameter;
+	if (count == 0)
+		return OutOfMemory;
+	if (count != pen->dash_count)
+		return InvalidParameter;
 
 	memcpy (dash, pen->dash_array, count * sizeof (float));
 
@@ -775,13 +850,13 @@ GdipSetPenDashArray (GpPen *pen, GDIPCONST float *dash, int count)
 {
 	float *dash_array;
 
-	g_return_val_if_fail (pen != NULL, InvalidParameter);
-	g_return_val_if_fail (dash != NULL, InvalidParameter);
-	g_return_val_if_fail (count > 0, InvalidParameter);
+	if (!pen || !dash || (count <= 0))
+		return InvalidParameter;
 
 	if (pen->dash_count != count || pen->own_dash_array == FALSE) {
 		dash_array = (float *) GdipAlloc (count * sizeof (float));
-		g_return_val_if_fail (dash_array != NULL, OutOfMemory);
+		if (!dash_array)
+			return OutOfMemory;
 
 		/* free the existing values, if we own them */
 		if (pen->dash_count != 0 && pen->own_dash_array == TRUE)
@@ -804,8 +879,8 @@ GdipSetPenDashArray (GpPen *pen, GDIPCONST float *dash, int count)
 GpStatus
 GdipGetPenCompoundCount (GpPen *pen, int *count)
 {
-	g_return_val_if_fail (pen != NULL, InvalidParameter);
-	g_return_val_if_fail (count != NULL, InvalidParameter);
+	if (!pen || !count)
+		return InvalidParameter;
 
 	*count = pen->compound_count;
 	return Ok;
@@ -816,13 +891,13 @@ GdipSetPenCompoundArray (GpPen *pen, const float *compound, int count)
 {
 	float *compound_array;
 
-	g_return_val_if_fail (pen != NULL, InvalidParameter);
-	g_return_val_if_fail (compound != NULL, InvalidParameter);
-	g_return_val_if_fail (count > 0, InvalidParameter);
+	if (!pen || !compound || (count <= 0))
+		return InvalidParameter;
 
 	if (pen->compound_count != count) {
 		compound_array = (float *) GdipAlloc (count * sizeof (float));
-		g_return_val_if_fail (compound_array != NULL, OutOfMemory);
+		if (!compound_array)
+			return OutOfMemory;
 
 		/* free the existing values */
 		if (pen->compound_count != 0)
@@ -839,9 +914,8 @@ GdipSetPenCompoundArray (GpPen *pen, const float *compound, int count)
 GpStatus
 GdipGetPenCompoundArray (GpPen *pen, float *compound, int count)
 {
-	g_return_val_if_fail (pen != NULL, InvalidParameter);
-	g_return_val_if_fail (compound != NULL, InvalidParameter);
-	g_return_val_if_fail (count == pen->compound_count, InvalidParameter);
+	if (!pen || !compound || (count != pen->compound_count))
+		return InvalidParameter;
 
 	memcpy (compound, pen->compound_array, count * sizeof (float));
 
@@ -851,7 +925,8 @@ GdipGetPenCompoundArray (GpPen *pen, float *compound, int count)
 GpStatus
 GdipSetPenStartCap (GpPen *pen, GpLineCap startCap)
 {
-	g_return_val_if_fail (pen != NULL, InvalidParameter);
+	if (!pen)
+		return InvalidParameter;
 
 	pen->line_cap = startCap;
 	pen->changed = TRUE;
@@ -861,18 +936,20 @@ GdipSetPenStartCap (GpPen *pen, GpLineCap startCap)
 GpStatus
 GdipGetPenStartCap (GpPen *pen, GpLineCap *startCap)
 {
-	g_return_val_if_fail (pen != NULL, InvalidParameter);
-	g_return_val_if_fail (startCap != NULL, InvalidParameter);
+	if (!pen || !startCap)
+		return InvalidParameter;
 
 	*startCap = pen->line_cap;
 
 	return Ok;
 }
 
+/* MonoTODO - ignored (always same as start cap) */
 GpStatus
 GdipSetPenEndCap (GpPen *pen, GpLineCap endCap)
 {
-	g_return_val_if_fail (pen != NULL, InvalidParameter);
+	if (!pen)
+		return InvalidParameter;
 
 	/*
 	 * This is currently ignored, as Cairo does not support
@@ -884,49 +961,84 @@ GdipSetPenEndCap (GpPen *pen, GpLineCap endCap)
 	return Ok;
 }
 
+/* MonoTODO - ignored (always same as start cap) */
 GpStatus
 GdipGetPenEndCap (GpPen *pen, GpLineCap *endCap)
 {
-	g_return_val_if_fail (pen != NULL, InvalidParameter);
-	g_return_val_if_fail (endCap != NULL, InvalidParameter);
+	if (!pen || !endCap)
+		return InvalidParameter;
 
 	*endCap = pen->end_cap;
 
 	return Ok;
 }
 
+/* MonoTODO - ignored */
 GpStatus
 GdipSetPenDashCap197819 (GpPen *pen, GpDashCap dashCap)
 {
-	return NotImplemented;
+	if (!pen)
+		return InvalidParameter;
+
+	pen->dash_cap = dashCap;
+	return Ok;
 }
 
+/* MonoTODO - ignored */
 GpStatus
 GdipGetPenDashCap197819 (GpPen *pen, GpDashCap *dashCap)
 {
-	return NotImplemented;
+	if (!pen || !dashCap)
+		return InvalidParameter;
+
+	*dashCap = pen->dash_cap;
+	return Ok;
 }
 
+/* MonoTODO - not implemented */
 GpStatus
 GdipSetPenCustomStartCap (GpPen *pen, GpCustomLineCap *customCap)
 {
-	return NotImplemented;
+	static bool called = FALSE;
+	if (!called) {
+		g_warning ("GdipSetPenCustomStartCap isn't implemented");
+		called = TRUE;
+	}
+	return Ok;
 }
 
+/* MonoTODO - not implemented */
 GpStatus
 GdipGetPenCustomStartCap (GpPen *pen, GpCustomLineCap **customCap)
 {
-	return NotImplemented;
+	static bool called = FALSE;
+	if (!called) {
+		g_warning ("GdipGetPenCustomStartCap isn't implemented");
+		called = TRUE;
+	}
+	return Ok;
 }
 
+/* MonoTODO - not implemented */
 GpStatus
 GdipSetPenCustomEndCap (GpPen *pen, GpCustomLineCap *customCap)
 {
-	return NotImplemented;
+	static bool called = FALSE;
+	if (!called) {
+		g_warning ("GdipSetPenCustomEndCap isn't implemented");
+		called = TRUE;
+	}
+	return Ok;
 }
 
+/* MonoTODO - not implemented */
 GpStatus
 GdipGetPenCustomEndCap (GpPen *pen, GpCustomLineCap **customCap)
 {
-	return NotImplemented;
+	static bool called = FALSE;
+	if (!called) {
+		g_warning ("GdipGetPenCustomEndCap isn't implemented");
+		called = TRUE;
+	}
+	return Ok;
 }
